# -*- coding: utf-8 -*-

from typing import Union, Sequence, Callable

import brainpy.math as bm
from brainpy.dyn.base import NeuGroup
from brainpy.initialize import ZeroInit, OneInit, Initializer, init_param
from brainpy.integrators.fde import CaputoL1Schema
from brainpy.integrators.fde import GLShortMemory
from brainpy.integrators.joint_eq import JointEq
from brainpy.tools.checking import check_float, check_integer
from brainpy.tools.checking import check_initializer
from brainpy.types import Shape, Tensor

__all__ = [
  'FractionalNeuron',
  'FractionalFHR',
  'FractionalIzhikevich',
]


class FractionalNeuron(NeuGroup):
  """Fractional-order neuron model."""
  pass


class FractionalFHR(FractionalNeuron):
  r"""The fractional-order FH-R model [1]_.

  FitzHugh and Rinzel introduced FH-R model (1976, in an unpublished article),
  which is the modification of the classical FHN neuron model. The fractional-order
  FH-R model is described as

  .. math::

     \begin{array}{rcl}
     \frac{{d}^{\alpha }v}{d{t}^{\alpha }} & = & v-{v}^{3}/3-w+y+I={f}_{1}(v,w,y),\\
     \frac{{d}^{\alpha }w}{d{t}^{\alpha }} & = & \delta (a+v-bw)={f}_{2}(v,w,y),\\
     \frac{{d}^{\alpha }y}{d{t}^{\alpha }} & = & \mu (c-v-dy)={f}_{3}(v,w,y),
     \end{array}

  where :math:`v, w` and :math:`y` represent the membrane voltage, recovery variable
  and slow modulation of the current respectively.
  :math:`I` measures the constant magnitude of external stimulus current, and :math:`\alpha`
  is the fractional exponent which ranges in the interval :math:`(0 < \alpha \le 1)`.
  :math:`a, b, c, d, \delta` and :math:`\mu` are the system parameters.

  The system reduces to the original classical order system when :math:`\alpha=1`.

  :math:`\mu` indicates a small parameter that determines the pace of the slow system
  variable :math:`y`. The fast subsystem (:math:`v-w`) presents a relaxation oscillator
  in the phase plane where :math:`\delta`  is a small parameter.
  :math:`v` is expressed in mV (millivolt) scale. Time :math:`t` is in ms (millisecond) scale.
  It exhibits tonic spiking or quiescent state depending on the parameter sets for a fixed
  value of :math:`I`. The parameter :math:`a` in the 2D FHN model corresponds to the
  parameter :math:`c` of the FH-R neuron model. If we decrease the value of :math:`a`,
  it causes longer intervals between two burstings, however there exists :math:`a`
  relatively fixed time of bursting duration. With the increasing of :math:`a`, the
  interburst intervals become shorter and periodic bursting changes to tonic spiking.

  Examples
  --------

  - [(Mondal, et, al., 2019): Fractional-order FitzHugh-Rinzel bursting neuron model](https://brainpy-examples.readthedocs.io/en/latest/neurons/2019_Fractional_order_FHR_model.html)


  Parameters
  ----------
  size: int, sequence of int
    The size of the neuron group.
  alpha: float, tensor
    The fractional order.
  num_memory: int
    The total number of the short memory.

  References
  ----------
  .. [1] Mondal, A., Sharma, S.K., Upadhyay, R.K. *et al.* Firing activities of a fractional-order FitzHugh-Rinzel bursting neuron model and its coupled dynamics. *Sci Rep* **9,** 15721 (2019). https://doi.org/10.1038/s41598-019-52061-4
  """

  def __init__(
      self,
      size: Shape,
      alpha: Union[float, Sequence[float]],
      num_memory: int = 1000,
      a: Union[float, Tensor, Initializer, Callable] = 0.7,
      b: Union[float, Tensor, Initializer, Callable] = 0.8,
      c: Union[float, Tensor, Initializer, Callable] = -0.775,
      d: Union[float, Tensor, Initializer, Callable] = 1.,
      delta: Union[float, Tensor, Initializer, Callable] = 0.08,
      mu: Union[float, Tensor, Initializer, Callable] = 0.0001,
      Vth: Union[float, Tensor, Initializer, Callable] = 1.8,
      V_initializer: Union[Initializer, Callable, Tensor] = OneInit(2.5),
      w_initializer: Union[Initializer, Callable, Tensor] = ZeroInit(),
      y_initializer: Union[Initializer, Callable, Tensor] = ZeroInit(),
      name: str = None
  ):
    super(FractionalFHR, self).__init__(size, name=name)

    # fractional order
    self.alpha = alpha
    check_integer(num_memory, 'num_memory', allow_none=False)

    # parameters
    self.a = init_param(a, self.num, allow_none=False)
    self.b = init_param(b, self.num, allow_none=False)
    self.c = init_param(c, self.num, allow_none=False)
    self.d = init_param(d, self.num, allow_none=False)
    self.mu = init_param(mu, self.num, allow_none=False)
    self.Vth = init_param(Vth, self.num, allow_none=False)
    self.delta = init_param(delta, self.num, allow_none=False)

    # variables
    check_initializer(V_initializer, 'V_initializer', allow_none=False)
    check_initializer(w_initializer, 'w_initializer', allow_none=False)
    check_initializer(y_initializer, 'y_initializer', allow_none=False)
    self.V = bm.Variable(init_param(V_initializer, (self.num,)))
    self.w = bm.Variable(init_param(w_initializer, (self.num,)))
    self.y = bm.Variable(init_param(y_initializer, (self.num,)))
    self.input = bm.Variable(bm.zeros(self.num))
    self.spike = bm.Variable(bm.zeros(self.num, dtype=bool))
    self.t_last_spike = bm.Variable(bm.ones(self.num) * -1e7)

    # integral function
    self.integral = GLShortMemory(self.derivative,
                                  alpha=alpha,
                                  num_memory=num_memory,
                                  inits=[self.V, self.w, self.y])

  def dV(self, V, t, w, y):
    return V - V ** 3 / 3 - w + y + self.input

  def dw(self, w, t, V):
    return self.delta * (self.a + V - self.b * w)

  def dy(self, y, t, V):
    return self.mu * (self.c - V - self.d * y)

  @property
  def derivative(self):
    return JointEq([self.dV, self.dw, self.dy])

  def update(self, _t, _dt):
    V, w, y = self.integral(self.V, self.w, self.y, _t, _dt)
    self.spike.value = bm.logical_and(V >= self.Vth, self.V < self.Vth)
    self.t_last_spike.value = bm.where(self.spike, _t, self.t_last_spike)
    self.V.value = V
    self.w.value = w
    self.y.value = y
    self.input[:] = 0.

  def set_init(self, values: dict):
    for k, v in values.items():
      if k not in self.integral.inits:
        raise ValueError(f'Variable "{k}" is not defined in this model.')
      variable = getattr(self, k)
      variable[:] = v
      self.integral.inits[k][:] = v


class FractionalIzhikevich(FractionalNeuron):
  r"""Fractional-order Izhikevich model [10]_.

  The fractional-order Izhikevich model is given by

  .. math::

     \begin{aligned}
      &\tau \frac{d^{\alpha} v}{d t^{\alpha}}=\mathrm{f} v^{2}+g v+h-u+R I \\
      &\tau \frac{d^{\alpha} u}{d t^{\alpha}}=a(b v-u)
      \end{aligned}

  where :math:`\alpha` is the fractional order (exponent) such that :math:`0<\alpha\le1`.
  It is a commensurate system that reduces to classical Izhikevich model at :math:`\alpha=1`.

  The time :math:`t` is in ms; and the system variable :math:`v` expressed in mV
  corresponds to membrane voltage. Moreover, :math:`u` expressed in mV is the
  recovery variable that corresponds to the activation of K+ ionic current and
  inactivation of Na+ ionic current.

  The parameters :math:`f, g, h` are fixed constants (should not be changed) such
  that :math:`f=0.04` (mV)−1, :math:`g=5, h=140` mV; and :math:`a` and :math:`b` are
  dimensionless parameters. The time constant :math:`\tau=1` ms; the resistance
  :math:`R=1` Ω; and :math:`I` expressed in mA measures the injected (applied)
  dc stimulus current to the system.

  When the membrane voltage reaches the spike peak :math:`v_{peak}`, the two variables
  are rest as follow:

  .. math::

     \text { if } v \geq v_{\text {peak }} \text { then }\left\{\begin{array}{l}
      v \leftarrow c \\
      u \leftarrow u+d
      \end{array}\right.

  we used :math:`v_{peak}=30` mV, and :math:`c` and :math:`d` are parameters expressed
  in mV. When the spike reaches its peak value, the membrane voltage :math:`v` and the
  recovery variable :math:`u` are reset according to the above condition.

  Examples
  --------

  - [(Teka, et. al, 2018): Fractional-order Izhikevich neuron model](https://brainpy-examples.readthedocs.io/en/latest/neurons/2018_Fractional_Izhikevich_model.html)


  References
  ----------
  .. [10] Teka, Wondimu W., Ranjit Kumar Upadhyay, and Argha Mondal. "Spiking and
          bursting patterns of fractional-order Izhikevich model." Communications
          in Nonlinear Science and Numerical Simulation 56 (2018): 161-176.

  """

  def __init__(
      self,
      size: Shape,
      alpha: Union[float, Sequence[float]],
      num_step: int,
      a: Union[float, Tensor, Initializer, Callable] = 0.02,
      b: Union[float, Tensor, Initializer, Callable] = 0.20,
      c: Union[float, Tensor, Initializer, Callable] = -65.,
      d: Union[float, Tensor, Initializer, Callable] = 8.,
      f: Union[float, Tensor, Initializer, Callable] = 0.04,
      g: Union[float, Tensor, Initializer, Callable] = 5.,
      h: Union[float, Tensor, Initializer, Callable] = 140.,
      R: Union[float, Tensor, Initializer, Callable] = 1.,
      tau: Union[float, Tensor, Initializer, Callable] = 1.,
      V_th: Union[float, Tensor, Initializer, Callable] = 30.,
      V_initializer: Union[Initializer, Callable, Tensor] = OneInit(-65.),
      u_initializer: Union[Initializer, Callable, Tensor] = OneInit(0.20 * -65.),
      name: str = None
  ):
    # initialization
    super(FractionalIzhikevich, self).__init__(size=size, name=name)

    # params
    self.alpha = alpha
    check_float(alpha, 'alpha', min_bound=0., max_bound=1., allow_none=False, allow_int=True)
    self.a = init_param(a, self.num, allow_none=False)
    self.b = init_param(b, self.num, allow_none=False)
    self.c = init_param(c, self.num, allow_none=False)
    self.d = init_param(d, self.num, allow_none=False)
    self.f = init_param(f, self.num, allow_none=False)
    self.g = init_param(g, self.num, allow_none=False)
    self.h = init_param(h, self.num, allow_none=False)
    self.tau = init_param(tau, self.num, allow_none=False)
    self.R = init_param(R, self.num, allow_none=False)
    self.V_th = init_param(V_th, self.num, allow_none=False)

    # variables
    check_initializer(V_initializer, 'V_initializer', allow_none=False)
    check_initializer(u_initializer, 'u_initializer', allow_none=False)
    self.V = bm.Variable(init_param(V_initializer, (self.num,)))
    self.u = bm.Variable(init_param(u_initializer, (self.num,)))
    self.input = bm.Variable(bm.zeros(self.num))
    self.spike = bm.Variable(bm.zeros(self.num, dtype=bool))
    self.t_last_spike = bm.Variable(bm.ones(self.num) * -1e7)

    # functions
    check_integer(num_step, 'num_step', allow_none=False)
    self.integral = CaputoL1Schema(f=self.derivative,
                                   alpha=alpha,
                                   num_step=num_step,
                                   inits=[self.V, self.u])

  def dV(self, V, t, u, I_ext):
    dVdt = self.f * V * V + self.g * V + self.h - u + self.R * I_ext
    return dVdt / self.tau

  def du(self, u, t, V):
    dudt = self.a * (self.b * V - u)
    return dudt / self.tau

  @property
  def derivative(self):
    return JointEq([self.dV, self.du])

  def update(self, _t, _dt):
    V, u = self.integral(self.V, self.u, t=_t, I_ext=self.input, dt=_dt)
    spikes = V >= self.V_th
    self.t_last_spike.value = bm.where(spikes, _t, self.t_last_spike)
    self.V.value = bm.where(spikes, self.c, V)
    self.u.value = bm.where(spikes, u + self.d, u)
    self.spike.value = spikes
    self.input[:] = 0.

  def set_init(self, values: dict):
    for k, v in values.items():
      if k not in self.integral.inits:
        raise ValueError(f'Variable "{k}" is not defined in this model.')
      variable = getattr(self, k)
      variable[:] = v
      self.integral.inits[k][:] = v



# -*- coding: utf-8 -*-

import jax.numpy as jnp

import brainpy.math as bm
from brainpy.errors import MathError
from brainpy.initialize import Initializer
from brainpy.nn.datatypes import MultipleData
from brainpy.nn.nodes.base.dense import Dense
from brainpy.tools.checking import check_shape_consistency

__all__ = [
  'LinearReadout',
]


class LinearReadout(Dense):
  """Linear readout node. Different from ``Dense``, this node has its own state.

  Parameters
  ----------
  num_unit: int
    The number of output features. A positive integer.
  weight_initializer: Initializer
    The weight initializer.
  bias_initializer: Optional, Initializer
    The bias initializer.
  trainable: bool
    Default is true.
  """
  data_pass = MultipleData('sequence')

  def __init__(self, num_unit: int, **kwargs):
    super(LinearReadout, self).__init__(num_unit=num_unit, **kwargs)

  def init_state(self, num_batch=1):
    return bm.zeros((num_batch,) + self.output_shape[1:], dtype=bm.float_)

  def forward(self, ff, fb=None, **shared_kwargs):
    h = super(LinearReadout, self).forward(ff, fb=fb, **shared_kwargs)
    self.state.value = h
    return h

  def online_init(self):
    _, free_shapes = check_shape_consistency(self.feedforward_shapes, -1, True)
    num_input = sum(free_shapes)
    if self.bias is not None:
      num_input += 1
    if self.feedback_shapes is not None:
      _, free_shapes = check_shape_consistency(self.feedback_shapes, -1, True)
      num_input += sum(free_shapes)
    self.online_fit_by.initialize(feature_in=num_input,
                                  feature_out=self.num_unit,
                                  name=self.name)

  def online_fit(self, target, ff, fb=None):
    if not isinstance(target, (bm.ndarray, jnp.ndarray)):
      raise MathError(f'"target" must be a tensor, but got {type(target)}')
    ff = bm.concatenate(ff, axis=-1)
    if ff.ndim != 2:
      raise ValueError(f'"ff" must be a 2D tensor with shape of (num_sample, '
                       f'num_feature), but we got {ff.shape}')
    if target.ndim != 2:
      raise ValueError(f'"target" must be a 2D tensor with shape of (num_sample, '
                       f'num_feature), but we got {target.shape}')
    if ff.shape[0] != target.shape[0]:
      raise ValueError(f'Batch size of the input and target data should be '
                       f'the same, while we got {ff.shape[0]} != {target.shape[0]}.')
    if target.shape[1] != self.state.shape[1]:
      raise MathError(f'The output dimension of output and target data should be '
                      f'the same, while we got {target.shape[1]} != {self.state.shape[1]}')
    if fb is not None:
      fb = bm.concatenate(fb, axis=-1)
      if fb.ndim != 2:
        raise ValueError(f'"fb" must be a 2D tensor with shape of (num_sample, '
                         f'num_feature), but we got {fb.shape}')
      if ff.shape[0] != fb.shape[0]:
        raise ValueError(f'Batch size of the feedforward and the feedback inputs should be '
                         f'the same, while we got {ff.shape[0]} != {fb.shape[0]}.')

    # data
    inputs = ff
    num_ff_input = ff.shape[1]
    if fb is not None:
      inputs = bm.concatenate([inputs, fb], axis=-1)
    if self.bias is not None:
      inputs = bm.concatenate([bm.ones((inputs.shape[0], 1)), inputs], axis=-1)

    # fitting
    dW = self.online_fit_by.call(target=target, input=inputs, output=self.state, name=self.name)

    # assign trained weights
    if self.bias is None:
      if fb is None:
        self.Wff += dW
      else:
        dWff, dWfb = bm.split(dW, [num_ff_input])
        self.Wff += dWff
        self.Wfb += dWfb
    else:
      if fb is None:
        db, dWff = bm.split(dW, [1])
        self.bias += db[0]
        self.Wff += dWff
      else:
        db, dWff, dWfb = bm.split(dW, [1, 1 + num_ff_input])
        self.bias += db[0]
        self.Wff += dWff
        self.Wfb += dWfb



# -*- coding: utf-8 -*-


import numpy as np

from brainpy import math as bm, tools
from brainpy.nn.base import Node
from brainpy.nn.datatypes import MultipleData
from brainpy.tools.checking import check_shape_consistency

__all__ = [
  'Concat', 'Select', 'Reshape', 'Summation',
]


class Concat(Node):
  """
  Concatenate multiple inputs into one.

  Parameters
  ----------
  axis : int
    The axis of concatenation to perform.
  """

  data_pass = MultipleData('sequence')

  def __init__(self, axis=-1, trainable=False, **kwargs):
    super(Concat, self).__init__(trainable=trainable, **kwargs)
    self.axis = axis

  def init_ff_conn(self):
    unique_shape, free_shapes = check_shape_consistency(self.feedforward_shapes, self.axis)
    out_size = list(unique_shape)
    out_size.insert(self.axis, sum(free_shapes))
    self.set_output_shape(out_size)

  def forward(self, ff, **shared_kwargs):
    return bm.concatenate(ff, axis=self.axis)


class Select(Node):
  """
  Select a subset of the given input.
  """

  def __init__(self, index, trainable=False, **kwargs):
    super(Select, self).__init__(trainable=trainable, **kwargs)
    if isinstance(index, int):
      self.index = bm.asarray([index]).value

  def init_ff_conn(self):
    out_size = bm.zeros(self.feedforward_shapes[1:])[self.index].shape
    self.set_output_shape((None,) + out_size)

  def forward(self, ff, **shared_kwargs):
    return ff[..., self.index]


class Reshape(Node):
  """
  Reshape the input tensor to another tensor.

  Parameters
  ----------
  shape: int, sequence of int
    The reshaped size. This shape does not contain the batch size.
  """

  def __init__(self, shape, trainable=False, **kwargs):
    super(Reshape, self).__init__(trainable=trainable, **kwargs)
    self.shape = tools.to_size(shape)
    assert (None not in self.shape), 'Batch size can not be defined in the reshaped size.'

  def init_ff_conn(self):
    in_size = self.feedforward_shapes[1:]
    if -1 in self.shape:
      assert self.shape.count(-1) == 1, f'Cannot set shape with multiple -1. But got {self.shape}'
      length = np.prod(in_size)
      out_size = list(self.shape)
      m1_idx = out_size.index(-1)
      other_shape = out_size[:m1_idx] + out_size[m1_idx + 1:]
      m1_length = int(length / np.prod(other_shape))
      out_size[m1_idx] = m1_length
    else:
      assert np.prod(in_size) == np.prod(self.shape)
      out_size = self.shape
    self.set_output_shape((None,) + tuple(out_size))

  def forward(self, ff, **shared_kwargs):
    return bm.reshape(ff, self.shape)


class Summation(Node):
  """
  Sum all input tensors into one.

  All inputs should be broadcast compatible.
  """
  data_pass = MultipleData('sequence')

  def __init__(self, trainable=False, **kwargs):
    super(Summation, self).__init__(trainable=trainable, **kwargs)

  def init_ff_conn(self):
    unique_shape, _ = check_shape_consistency(self.feedforward_shapes, None, True)
    self.set_output_shape(list(unique_shape))

  def forward(self, ff, **shared_kwargs):
    res = ff[0]
    for v in ff[1:]:
      res = res + v
    return res



# -*- coding: utf-8 -*-

from typing import Dict, Sequence, Union, Callable

import tqdm.auto
from jax.experimental.host_callback import id_tap
from jax.tree_util import tree_map

from brainpy.base import Base
import brainpy.math as bm
from brainpy.errors import NoImplementationError
from brainpy.nn.algorithms.online import get, OnlineAlgorithm, RLS
from brainpy.nn.base import Node
from brainpy.nn.utils import (serialize_kwargs,
                              check_data_batch_size,
                              check_rnn_data_time_step)
from brainpy.types import Tensor
from .rnn_trainer import RNNTrainer

__all__ = [
  'OnlineTrainer',
  'ForceTrainer',
]


class OnlineTrainer(RNNTrainer):
  """Online trainer for models with recurrent dynamics.

  Parameters
  ----------
  target: Node
    The target model to train.
  fit_method: OnlineAlgorithm, Callable, dict, str
    The fitting method applied to the target model.
    - It can be a string, which specify the shortcut name of the training algorithm.
      Like, ``fit_method='ridge'`` means using the RLS method.
      All supported fitting methods can be obtained through
      :py:func:`brainpy.nn.runners.get_supported_online_methods`
    - It can be a dict, whose "name" item specifies the name of the training algorithm,
      and the others parameters specify the initialization parameters of the algorithm.
      For example, ``fit_method={'name': 'ridge', 'beta': 1e-4}``.
    - It can be an instance of :py:class:`brainpy.nn.runners.OnlineAlgorithm`.
      For example, ``fit_meth=bp.nn.runners.RLS(alpha=1e-5)``.
    - It can also be a callable function.
  **kwargs
    The other general parameters for RNN running initialization.
  """

  def __init__(
      self,
      target: Node,
      fit_method: Union[OnlineAlgorithm, Callable, Dict, str] = None,
      **kwargs
  ):
    super(OnlineTrainer, self).__init__(target=target, **kwargs)

    # training method
    if fit_method is None:
      fit_method = RLS(alpha=1e-7)
    elif isinstance(fit_method, str):
      fit_method = get(fit_method)()
    elif isinstance(fit_method, dict):
      name = fit_method.pop('name')
      fit_method = get(name)(**fit_method)
    self.fit_method = fit_method
    if not callable(fit_method):
      raise ValueError(f'"train_method" must be an instance of callable function, '
                       f'but we got {type(fit_method)}.')

    # check the required interface in the trainable nodes
    self._check_interface()

    # set the training method
    for node in self.train_nodes:
      node.online_fit_by = fit_method

    # initialize the fitting method
    for node in self.train_nodes:
      node.online_init()

    # update dynamical variables
    if isinstance(self.fit_method, Base):
      self.dyn_vars.update(self.fit_method.vars().unique())

    # training function
    self._f_train = dict()

  def __repr__(self):
    name = self.__class__.__name__
    prefix = ' ' * len(name)
    return (f'{name}(target={self.target}, \n\t'
            f'{prefix}jit={self.jit}, \n\t'
            f'{prefix}fit_method={self.fit_method})')

  def fit(
      self,
      train_data: Sequence,
      test_data=None,
      reset: bool = False,
      shared_kwargs: Dict = None,
      forced_states: Dict[str, Tensor] = None,
      forced_feedbacks: Dict[str, Tensor] = None,
      initial_states: Dict[str, Tensor] = None,
      initial_feedbacks: Dict[str, Tensor] = None,
  ):
    # checking training and testing data
    if not isinstance(train_data, (list, tuple)):
      raise ValueError(f"{self.__class__.__name__} only support "
                       f"training data with the format of (X, Y) pair, "
                       f"but we got a {type(train_data)}.")
    if len(train_data) != 2:
      raise ValueError(f"{self.__class__.__name__} only support "
                       f"training data with the format of (X, Y) pair, "
                       f"but we got a sequence with length {len(train_data)}")
    if test_data is not None:
      raise ValueError(f'{self.__class__.__name__} does not support testing data.')
    xs, ys = train_data

    # format input data
    xs, num_step, num_batch = self._check_xs(xs, move_axis=True)

    # format target data
    ys = self._check_ys(ys, num_batch=num_batch, num_step=num_step, move_axis=True)

    # set initial states
    self._set_initial_states(initial_states)
    self._set_initial_feedbacks(initial_feedbacks)

    # get forced data
    forced_states = self._check_forced_states(forced_states, num_batch, num_step)
    forced_feedbacks = self._check_forced_feedbacks(forced_feedbacks, num_batch, num_step)

    # reset the model states
    if reset:
      self.target.initialize(num_batch)

    # init monitor
    for key in self.mon.item_contents.keys():
      self.mon.item_contents[key] = []  # reshape the monitor items

    # init progress bar
    if self.progress_bar:
      if num_step is None:
        num_step = check_rnn_data_time_step(xs)
      self._pbar = tqdm.auto.tqdm(total=num_step)
      self._pbar.set_description(f"Train {num_step} steps: ", refresh=True)

    # prediction
    hists = self._fit(xs=xs,
                      ys=ys,
                      forced_states=forced_states,
                      forced_feedbacks=forced_feedbacks,
                      shared_kwargs=shared_kwargs)

    # close the progress bar
    if self.progress_bar:
      self._pbar.close()

    # post-running for monitors
    for key in self.mon.item_names:
      self.mon.item_contents[key] = hists[key]
    if self.numpy_mon_after_run:
      self.mon.numpy()

  def _fit(
      self,
      xs: Dict[str, Tensor],
      ys: Dict[str, Tensor],
      shared_kwargs: Dict = None,
      forced_states: Dict[str, Tensor] = None,
      forced_feedbacks: Dict[str, Tensor] = None,
  ):
    """Predict the output according to the inputs.

    Parameters
    ----------
    xs: dict
      Each tensor should have the shape of `(num_time, num_batch, num_feature)`.
    ys: dict
      Each tensor should have the shape of `(num_time, num_batch, num_feature)`.
    forced_states: dict
      The forced state values.
    forced_feedbacks: dict
      The forced feedback output values.
    shared_kwargs: optional, dict
      The shared keyword arguments.

    Returns
    -------
    outputs, hists
      A tuple of pair of (outputs, hists).
    """
    _predict_func = self._get_fit_func(shared_kwargs)
    # rune the model
    forced_states = dict() if forced_states is None else forced_states
    forced_feedbacks = dict() if forced_feedbacks is None else forced_feedbacks
    hists = _predict_func([xs, ys, forced_states, forced_feedbacks])
    f1 = lambda x: bm.moveaxis(x, 0, 1)
    f2 = lambda x: isinstance(x, bm.JaxArray)
    hists = tree_map(f1, hists, is_leaf=f2)
    return hists

  def _get_fit_func(self, shared_kwargs: Dict = None):
    if shared_kwargs is None: shared_kwargs = dict()
    shared_kwargs_str = serialize_kwargs(shared_kwargs)
    if shared_kwargs_str not in self._f_train:
      self._f_train[shared_kwargs_str] = self._make_fit_func(shared_kwargs)
    return self._f_train[shared_kwargs_str]

  def _make_fit_func(self, shared_kwargs: Dict):
    if not isinstance(shared_kwargs, dict):
      raise ValueError(f'"shared_kwargs" must be a dict, '
                       f'but got {type(shared_kwargs)}')
    add_monitors = self._add_monitor_items()

    def _step_func(all_inputs):
      xs, ys, forced_states, forced_feedbacks = all_inputs
      monitors = tuple(self.mon.item_contents.keys())

      _, outs = self.target(xs,
                            forced_states=forced_states,
                            forced_feedbacks=forced_feedbacks,
                            monitors=monitors + add_monitors,
                            **shared_kwargs)
      for node in self.train_nodes:
        ff = outs[f'{node.name}.inputs']
        fb = outs[f'{node.name}.feedbacks']
        target = ys[node.name]
        node.online_fit(target, ff, fb=fb)
      for key in add_monitors:
        outs.pop(key)

      if self.progress_bar and (self._pbar is not None):
        id_tap(lambda *args: self._pbar.update(), ())
      return outs

    if self.jit['fit']:
      dyn_vars = self.target.vars()
      dyn_vars.update(self.dyn_vars)
      f = bm.make_loop(_step_func, dyn_vars=dyn_vars.unique(), has_return=True)
      return lambda all_inputs: f(all_inputs)[1]

    else:
      def run_func(all_inputs):
        xs, ys, forced_states, forced_feedbacks = all_inputs
        monitors = {key: [] for key in self.mon.item_contents.keys()}
        num_step = check_data_batch_size(xs)
        for i in range(num_step):
          one_xs = {key: tensor[i] for key, tensor in xs.items()}
          one_ys = {key: tensor[i] for key, tensor in ys.items()}
          one_forced_states = {key: tensor[i] for key, tensor in forced_states.items()}
          one_forced_feedbacks = {key: tensor[i] for key, tensor in forced_feedbacks.items()}
          mon = _step_func([one_xs, one_ys, one_forced_states, one_forced_feedbacks])
          for key, value in mon.items():
            monitors[key].append(value)
        for key, value in monitors.items():
          monitors[key] = bm.asarray(value)
        return monitors
    return run_func

  def _add_monitor_items(self):
    added_items = set()
    for node in self.train_nodes:
      if f'{node.name}.inputs' not in self.mon.item_names:
        added_items.add(f'{node.name}.inputs')
      if f'{node.name}.feedbacks' not in self.mon.item_names:
        added_items.add(f'{node.name}.feedbacks')
    return tuple(added_items)

  def _check_interface(self):
    for node in self.train_nodes:
      if hasattr(node.online_fit, 'not_implemented'):
        if node.online_fit.not_implemented:
          raise NoImplementationError(
            f'The node \n\n{node}\n\n'
            f'is set to be trainable with {self.__class__.__name__} method. '
            f'However, it does not implement the required training '
            f'interface "online_fit()" function. '
          )
      if hasattr(node.online_init, 'not_implemented'):
        if node.online_init.not_implemented:
          raise NoImplementationError(
            f'The node \n\n{node}\n\n'
            f'is set to be trainable with {self.__class__.__name__} method. '
            f'However, it does not implement the required training '
            f'interface "online_init()" function. '
          )


class ForceTrainer(OnlineTrainer):
  """Force learning."""

  def __init__(self, target, alpha=1., **kwargs):
    fit_method = RLS(alpha=alpha)
    super(ForceTrainer, self).__init__(target=target,
                                       fit_method=fit_method,
                                       **kwargs)



# -*- coding: utf-8 -*-

"""Neural Networks (nn)"""

from .base import *
from .datatypes import *
from .graph_flow import *
from .nodes import *
from .graph_flow import *
from .operations import *
from .utils import *
from .runners import *
from . import algorithms




# -*- coding: utf-8 -*-


__all__ = [
  'is_checking',
  'turn_on',
  'turn_off',
]

_check = True


def is_checking():
  """Whether the checking is turn on."""
  return _check


def turn_on():
  """Turn on the checking."""
  global _check
  _check = True


def turn_off():
  """Turn off the checking."""
  global _check
  _check = False



# -*- coding: utf-8 -*-


import jax
import jax.numpy as jnp
from contextlib import contextmanager

math = None

__all__ = [
  'Collector',
  'TensorCollector',
]


class Collector(dict):
  """A Collector is a dictionary (name, var) with some additional methods to make manipulation
  of collections of variables easy. A Collector is ordered by insertion order. It is the object
  returned by Base.vars() and used as input in many Collector instance: optimizers, jit, etc..."""

  def __setitem__(self, key, value):
    """Overload bracket assignment to catch potential conflicts during assignment."""
    if key in self:
      if id(self[key]) != id(value):
        raise ValueError(f'Name "{key}" conflicts: same name for {value} and {self[key]}.')
    dict.__setitem__(self, key, value)

  def replace(self, key, new_value):
    """Replace the original key with the new value."""
    self.pop(key)
    self[key] = new_value
    # dict.__setitem__(self, key, new_value)

  def update(self, other, **kwargs):
    assert isinstance(other, dict)
    for key, value in other.items():
      self[key] = value
    for key, value in kwargs.items():
      self[key] = value

  def __add__(self, other):
    gather = type(self)(self)
    gather.update(other)
    return gather

  def subset(self, var_type):
    """Get the subset of the (key, value) pair.

    ``subset()`` can be used to get a subset of some class:

    >>> import brainpy as bp
    >>>
    >>> some_collector = Collector()
    >>>
    >>> # get all trainable variables
    >>> some_collector.subset(bp.math.TrainVar)
    >>>
    >>> # get all JaxArray
    >>> some_collector.subset(bp.math.Variable)

    or, it can be used to get a subset of integrators:

    >>> # get all ODE integrators
    >>> some_collector.subset(bp.ode.ODEIntegrator)

    Parameters
    ----------
    var_type : type
      The type/class to match.
    """
    gather = type(self)()
    for key, value in self.items():
      if isinstance(value, var_type):
        gather[key] = value
    return gather

  def unique(self):
    """Get a new type of collector with unique values.

    If one value is assigned to two or more keys,
    then only one pair of (key, value) will be returned.
    """
    gather = type(self)()
    seen = set()
    for k, v in self.items():
      if id(v) not in seen:
        seen.add(id(v))
        gather[k] = v
    return gather


class TensorCollector(Collector):
  """A ArrayCollector is a dictionary (name, var)
  with some additional methods to make manipulation
  of collections of variables easy. A Collection
  is ordered by insertion order. It is the object
  returned by DynamicalSystem.vars() and used as input
  in many DynamicalSystem instance: optimizers, Jit, etc..."""

  def __setitem__(self, key, value):
    """Overload bracket assignment to catch potential conflicts during assignment."""
    global math
    if math is None: from brainpy import math

    assert isinstance(value, math.ndarray)
    if key in self:
      if id(self[key]) != id(value):
        raise ValueError(f'Name "{key}" conflicts: same name for {value} and {self[key]}.')
    dict.__setitem__(self, key, value)

  def assign(self, inputs):
    """Assign data to all values.

    Parameters
    ----------
    inputs : dict
      The data for each value in this collector.
    """
    if len(self) != len(inputs):
      raise ValueError(f'The target has {len(inputs)} data, while we got '
                       f'an input value with the length of {len(inputs)}.')
    for key, v in self.items():
      v.value = inputs[key]

  def dict(self):
    """Get a dict with the key and the value data.
    """
    gather = dict()
    for k, v in self.items():
      gather[k] = v.value
    return gather

  def data(self):
    """Get all data in each value."""
    return [x.value for x in self.values()]

  # @contextmanager
  # def replicate(self):
  #   """A context manager to use in a with statement that replicates
  #   the variables in this collection to multiple devices.
  #
  #   Important: replicating also updates the random state in order
  #   to have a new one per device.
  #   """
  #   global math
  #   if math is None: from brainpy import math
  #
  #   replicated, saved_states = {}, {}
  #   x = jnp.zeros((jax.local_device_count(), 1), dtype=math.float_)
  #   sharded_x = jax.pmap(lambda x: x, axis_name='device')(x)
  #   devices = [b.device() for b in sharded_x.device_buffers]
  #   num_device = len(devices)
  #   for k, d in self.items():
  #     if isinstance(d, math.random.RandomState):
  #       replicated[k] = jax.device_put_sharded([shard for shard in d.split(num_device)], devices)
  #       saved_states[k] = d.value
  #     else:
  #       replicated[k] = jax.device_put_replicated(d.value, devices)
  #   self.assign(replicated)
  #   yield
  #   visited = set()
  #   for k, d in self.items():
  #     # Careful not to reduce twice in case of
  #     # a variable and a reference to it.
  #     if id(d) not in visited:
  #       if isinstance(d, math.random.RandomState):
  #         d.value = saved_states[k]
  #       else:
  #         d.value = reduce_func(d)
  #       visited.add(id(d))



# -*- coding: utf-8 -*-

"""
Numerical methods for ordinary differential equations (ODEs).
"""

from .base import *
from .generic import *
from .adaptive_rk import *
from .explicit_rk import *
from .exponential import *



# -*- coding: utf-8 -*-

import math as pm
import warnings
from typing import Union, Dict, Callable

import jax.numpy as jnp
import numpy as np

import brainpy.math as bm
from brainpy import tools
from brainpy.base.base import Base
from brainpy.base.collector import Collector
from brainpy.connect import TwoEndConnector, MatConn, IJConn
from brainpy.initialize import Initializer, ZeroInit, init_param
from brainpy.errors import ModelBuildError
from brainpy.integrators.base import Integrator
from brainpy.types import Tensor

__all__ = [
  'DynamicalSystem',
  'Container',
  'Network',
  'ConstantDelay',
  'NeuGroup',
  'TwoEndConn',
]

_error_msg = 'Unknown type of the update function: {} ({}). ' \
             'Currently, BrainPy only supports: \n' \
             '1. function \n' \
             '2. function name (str) \n' \
             '3. tuple/dict of functions \n' \
             '4. tuple of function names \n'


class DynamicalSystem(Base):
  """Base Dynamical System class.

  Any object has step functions will be a dynamical system.
  That is to say, in BrainPy, the essence of the dynamical system
  is the "step functions".

  Parameters
  ----------
  name : str, optional
      The name of the dynamic system.
  """

  def __init__(self, name=None):
    super(DynamicalSystem, self).__init__(name=name)

  @property
  def steps(self):
    warnings.warn('.steps has been deprecated since version 2.0.3.', DeprecationWarning)
    return {}

  def ints(self, method='absolute'):
    """Collect all integrators in this node and the children nodes.

    Parameters
    ----------
    method : str
      The method to access the integrators.

    Returns
    -------
    collector : Collector
      The collection contained (the path, the integrator).
    """
    nodes = self.nodes(method=method)
    gather = Collector()
    for node_path, node in nodes.items():
      for k in dir(node):
        v = getattr(node, k)
        if isinstance(v, Integrator):
          gather[f'{node_path}.{k}' if node_path else k] = v
    return gather

  def __call__(self, *args, **kwargs):
    """The shortcut to call ``update`` methods."""
    return self.update(*args, **kwargs)

  def update(self, _t, _dt):
    """The function to specify the updating rule.
    Assume any dynamical system depends on the time variable ``t`` and
    the time step ``dt``.
    """
    raise NotImplementedError('Must implement "update" function by user self.')


class Container(DynamicalSystem):
  """Container object which is designed to add other instances of DynamicalSystem.

  Parameters
  ----------
  steps : tuple of function, tuple of str, dict of (str, function), optional
      The step functions.
  monitors : tuple, list, Monitor, optional
      The monitor object.
  name : str, optional
      The object name.
  show_code : bool
      Whether show the formatted code.
  ds_dict : dict of (str, )
      The instance of DynamicalSystem with the format of "key=dynamic_system".
  """

  def __init__(self, *ds_tuple, name=None, **ds_dict):
    super(Container, self).__init__(name=name)

    # children dynamical systems
    self.implicit_nodes = Collector()
    for ds in ds_tuple:
      if not isinstance(ds, DynamicalSystem):
        raise ModelBuildError(f'{self.__class__.__name__} receives instances of '
                              f'DynamicalSystem, however, we got {type(ds)}.')
      if ds.name in self.implicit_nodes:
        raise ValueError(f'{ds.name} has been paired with {ds}. Please change a unique name.')
    self.register_implicit_nodes({node.name: node for node in ds_tuple})
    for key, ds in ds_dict.items():
      if not isinstance(ds, DynamicalSystem):
        raise ModelBuildError(f'{self.__class__.__name__} receives instances of '
                              f'DynamicalSystem, however, we got {type(ds)}.')
      if key in self.implicit_nodes:
        raise ValueError(f'{key} has been paired with {ds}. Please change a unique name.')
    self.register_implicit_nodes(ds_dict)

  def update(self, _t, _dt):
    """Step function of a network.

    In this update function, the update functions in children systems are
    iteratively called.
    """
    nodes = self.nodes(level=1, include_self=False).subset(DynamicalSystem).unique()
    for node in nodes.values():
      node.update(_t, _dt)

  def __getattr__(self, item):
    child_ds = super(Container, self).__getattribute__('implicit_nodes')
    if item in child_ds:
      return child_ds[item]
    else:
      return super(Container, self).__getattribute__(item)


class Network(Container):
  """Base class to model network objects, an alias of Container.

  Network instantiates a network, which is aimed to load
  neurons, synapses, and other brain objects.

  Parameters
  ----------
  name : str, Optional
    The network name.
  monitors : optional, list of str, tuple of str
    The items to monitor.
  ds_tuple :
    A list/tuple container of dynamical system.
  ds_dict :
    A dict container of dynamical system.
  """

  def __init__(self, *ds_tuple, name=None, **ds_dict):
    super(Network, self).__init__(*ds_tuple, name=name, **ds_dict)


class ConstantDelay(DynamicalSystem):
  """Class used to model constant delay variables.

  This class automatically supports batch size on the last axis. For example, if
  you run batch with the size of (10, 100), where `100` are batch size, then this
  class can automatically support your batched data.
  For examples,

  >>> import brainpy as bp
  >>> bp.dyn.ConstantDelay(size=(10, 100), delay=10.)

  This class also support nonuniform delays.

  >>> bp.dyn.ConstantDelay(size=100, delay=bp.math.random.random(100) * 4 + 10)

  Parameters
  ----------
  size : int, list of int, tuple of int
    The delay data size.
  delay : int, float, function, ndarray
    The delay time. With the unit of `dt`.
  dt: float, optional
    The time precision.
  name : optional, str
    The name of the dynamic system.
  """

  def __init__(self, size, delay, dtype=None, dt=None, **kwargs):
    # dt
    self.dt = bm.get_dt() if dt is None else dt

    # data size
    if isinstance(size, int): size = (size,)
    if not isinstance(size, (tuple, list)):
      raise ModelBuildError(f'"size" must a tuple/list of int, but we got {type(size)}: {size}')
    self.size = tuple(size)

    # delay time length
    self.delay = delay

    # data and operations
    if isinstance(delay, (int, float)):  # uniform delay
      self.uniform_delay = True
      self.num_step = int(pm.ceil(delay / self.dt)) + 1
      self.out_idx = bm.Variable(bm.array([0], dtype=bm.uint32))
      self.in_idx = bm.Variable(bm.array([self.num_step - 1], dtype=bm.uint32))
      self.data = bm.Variable(bm.zeros((self.num_step,) + self.size, dtype=dtype))

    else:  # non-uniform delay
      self.uniform_delay = False
      if not len(self.size) == 1:
        raise NotImplementedError(f'Currently, BrainPy only supports 1D heterogeneous '
                                  f'delays, while we got the heterogeneous delay with '
                                  f'{len(self.size)}-dimensions.')
      self.num = tools.size2num(size)
      if bm.ndim(delay) != 1:
        raise ModelBuildError(f'Only support a 1D non-uniform delay. '
                              f'But we got {delay.ndim}D: {delay}')
      if delay.shape[0] != self.size[0]:
        raise ModelBuildError(f"The first shape of the delay time size must "
                              f"be the same with the delay data size. But "
                              f"we got {delay.shape[0]} != {self.size[0]}")
      delay = bm.around(delay / self.dt)
      self.diag = bm.array(bm.arange(self.num), dtype=bm.int_)
      self.num_step = bm.array(delay, dtype=bm.uint32) + 1
      self.in_idx = bm.Variable(self.num_step - 1)
      self.out_idx = bm.Variable(bm.zeros(self.num, dtype=bm.uint32))
      self.data = bm.Variable(bm.zeros((self.num_step.max(),) + size, dtype=dtype))

    super(ConstantDelay, self).__init__(**kwargs)

  @property
  def oldest(self):
    return self.pull()

  @property
  def latest(self):
    if self.uniform_delay:
      return self.data[self.in_idx[0]]
    else:
      return self.data[self.in_idx, self.diag]

  def pull(self):
    if self.uniform_delay:
      return self.data[self.out_idx[0]]
    else:
      return self.data[self.out_idx, self.diag]

  def push(self, value):
    if self.uniform_delay:
      self.data[self.in_idx[0]] = value
    else:
      self.data[self.in_idx, self.diag] = value

  def update(self, _t=None, _dt=None, **kwargs):
    """Update the delay index."""
    self.in_idx[:] = (self.in_idx + 1) % self.num_step
    self.out_idx[:] = (self.out_idx + 1) % self.num_step

  def reset(self):
    """Reset the variables."""
    self.in_idx[:] = self.num_step - 1
    self.out_idx[:] = 0
    self.data[:] = 0


class NeuGroup(DynamicalSystem):
  """Base class to model neuronal groups.

  There are several essential attributes:

  - ``size``: the geometry of the neuron group. For example, `(10, )` denotes a line of
    neurons, `(10, 10)` denotes a neuron group aligned in a 2D space, `(10, 15, 4)` denotes
    a 3-dimensional neuron group.
  - ``num``: the flattened number of neurons in the group. For example, `size=(10, )` => \
    `num=10`, `size=(10, 10)` => `num=100`, `size=(10, 15, 4)` => `num=600`.

  Parameters
  ----------
  size : int, tuple of int, list of int
    The neuron group geometry.
  name : optional, str
    The name of the dynamic system.
  """

  def __init__(self, size, name=None):
    # size
    if isinstance(size, (list, tuple)):
      if len(size) <= 0:
        raise ModelBuildError('size must be int, or a tuple/list of int.')
      if not isinstance(size[0], int):
        raise ModelBuildError('size must be int, or a tuple/list of int.')
      size = tuple(size)
    elif isinstance(size, int):
      size = (size,)
    else:
      raise ModelBuildError('size must be int, or a tuple/list of int.')
    self.size = size
    self.num = tools.size2num(size)

    # initialize
    super(NeuGroup, self).__init__(name=name)

  def update(self, _t, _dt):
    """The function to specify the updating rule.

    Parameters
    ----------
    _t : float
      The current time.
    _dt : float
      The time step.
    """
    raise NotImplementedError(f'Subclass of {self.__class__.__name__} must '
                              f'implement "update" function.')


class TwoEndConn(DynamicalSystem):
  """Base class to model two-end synaptic connections.

  Parameters
  ----------
  pre : NeuGroup
      Pre-synaptic neuron group.
  post : NeuGroup
      Post-synaptic neuron group.
  conn : optional, ndarray, JaxArray, dict, TwoEndConnector
      The connection method between pre- and post-synaptic groups.
  name : str, optional
      The name of the dynamic system.
  """

  """Global delay variables. Useful when the same target
    variable is used in multiple mappings."""
  global_delay_vars: Dict[str, bm.LengthDelay] = dict()

  def __init__(
      self,
      pre: NeuGroup,
      post: NeuGroup,
      conn: Union[TwoEndConnector, Tensor, Dict[str, Tensor]] = None,
      name: str = None
  ):
    # local delay variables
    self.local_delay_vars: Dict[str, bm.LengthDelay] = dict()

    # pre or post neuron group
    # ------------------------
    if not isinstance(pre, NeuGroup):
      raise ModelBuildError('"pre" must be an instance of NeuGroup.')
    if not isinstance(post, NeuGroup):
      raise ModelBuildError('"post" must be an instance of NeuGroup.')
    self.pre = pre
    self.post = post

    # connectivity
    # ------------
    if isinstance(conn, TwoEndConnector):
      self.conn = conn(pre.size, post.size)
    elif isinstance(conn, (bm.ndarray, np.ndarray, jnp.ndarray)):
      if (pre.num, post.num) != conn.shape:
        raise ModelBuildError(f'"conn" is provided as a matrix, and it is expected '
                              f'to be an array with shape of (pre.num, post.num) = '
                              f'{(pre.num, post.num)}, however we got {conn.shape}')
      self.conn = MatConn(conn_mat=conn)
    elif isinstance(conn, dict):
      if not ('i' in conn and 'j' in conn):
        raise ModelBuildError(f'"conn" is provided as a dict, and it is expected to '
                              f'be a dictionary with "i" and "j" specification, '
                              f'however we got {conn}')
      self.conn = IJConn(i=conn['i'], j=conn['j'])
    elif isinstance(conn, str):
      self.conn = conn
    elif conn is None:
      self.conn = None
    else:
      raise ModelBuildError(f'Unknown "conn" type: {conn}')

    # initialize
    # ----------
    super(TwoEndConn, self).__init__(name=name)

  def check_pre_attrs(self, *attrs):
    """Check whether pre group satisfies the requirement."""
    if not hasattr(self, 'pre'):
      raise ModelBuildError('Please call __init__ function first.')
    for attr in attrs:
      if not isinstance(attr, str):
        raise ValueError(f'Must be string. But got {attr}.')
      if not hasattr(self.pre, attr):
        raise ModelBuildError(f'{self} need "pre" neuron group has attribute "{attr}".')

  def check_post_attrs(self, *attrs):
    """Check whether post group satisfies the requirement."""
    if not hasattr(self, 'post'):
      raise ModelBuildError('Please call __init__ function first.')
    for attr in attrs:
      if not isinstance(attr, str):
        raise ValueError(f'Must be string. But got {attr}.')
      if not hasattr(self.post, attr):
        raise ModelBuildError(f'{self} need "pre" neuron group has attribute "{attr}".')

  def register_delay(
      self,
      name: str,
      delay_step: Union[int, bm.ndarray, jnp.ndarray, Callable, Initializer],
      delay_target: Union[bm.JaxArray, jnp.ndarray],
      initial_delay_data: Union[Initializer, Callable] = None,
      domain: str = 'global'
  ):
    """Register delay variable.

    Parameters
    ----------
    name: str
      The delay variable name.
    delay_step: int, JaxArray, ndarray, callable, Initializer
      The number of the steps of the delay.
    delay_target: JaxArray, ndarray, Variable
      The target for delay.
    initial_delay_data: float, int, JaxArray, ndarray, callable, Initializer
      The initializer for the delay data.
    domain: str
      The domain of the delay data to store.

    Returns
    -------
    delay_step: int, JaxArray, ndarray
      The number of the delay steps.
    """
    # delay steps
    if delay_step is None:
      return delay_step
    elif isinstance(delay_step, int):
      delay_type = 'homo'
    elif isinstance(delay_step, (bm.ndarray, jnp.ndarray, np.ndarray)):
      delay_type = 'heter'
      delay_step = bm.asarray(delay_step)
    elif callable(delay_step):
      delay_step = init_param(delay_step, delay_target.shape, allow_none=False)
      delay_type = 'heter'
    else:
      raise ValueError(f'Unknown "delay_steps" type {type(delay_step)}, only support '
                       f'integer, array of integers, callable function, brainpy.init.Initializer.')
    if delay_type == 'heter':
      if delay_step.dtype not in [bm.int32, bm.int64]:
        raise ValueError('Only support delay steps of int32, int64. If your '
                         'provide delay time length, please divide the "dt" '
                         'then provide us the number of delay steps.')
      if delay_target.shape[0] != delay_step.shape[0]:
        raise ValueError(f'Shape is mismatched: {delay_target.shape[0]} != {delay_step.shape[0]}')
    max_delay_step = int(bm.max(delay_step))

    # delay domain
    if domain not in ['global', 'local']:
      raise ValueError('"domain" must be a string in ["global", "local"]. '
                       f'Bug we got {domain}.')

    # delay variable
    if domain == 'local':
      self.local_delay_vars[name] = bm.LengthDelay(delay_target, max_delay_step, initial_delay_data)
      self.register_implicit_nodes(self.local_delay_vars)
    else:
      if name not in self.global_delay_vars:
        self.global_delay_vars[name] = bm.LengthDelay(delay_target, max_delay_step, initial_delay_data)
        # save into local delay vars when first seen "var",
        # for later update current value!
        self.local_delay_vars[name] = self.global_delay_vars[name]
      else:
        if self.global_delay_vars[name].num_delay_step - 1 < max_delay_step:
          self.global_delay_vars[name].init(delay_target, max_delay_step, initial_delay_data)
      self.register_implicit_nodes(self.global_delay_vars)
    return delay_step

  def get_delay(
      self,
      name: str,
      delay_step: Union[int, bm.JaxArray, bm.ndarray],
      indices=None,
  ):
    """Get delay data according to the provided delay steps.

    Parameters
    ----------
    name: str
      The delay variable name.
    delay_step: int, JaxArray, ndarray
      The delay length.
    indices: optional, JaxArray, ndarray
      The indices of the delay.

    Returns
    -------
    delay_data: JaxArray, ndarray
      The delay data at the given time.
    """
    if name in self.global_delay_vars:
      if isinstance(delay_step, int):
        return self.global_delay_vars[name](delay_step, indices)
      else:
        if indices is None:
          indices = jnp.arange(delay_step.size)
        return self.global_delay_vars[name](delay_step, indices)
    elif name in self.local_delay_vars:
      if isinstance(delay_step, int):
        return self.local_delay_vars[name](delay_step)
      else:
        if indices is None:
          indices = jnp.arange(delay_step.size)
        return self.local_delay_vars[name](delay_step, indices)
    else:
      raise ValueError(f'{name} is not defined in delay variables.')

  def update_delay(
      self,
      name: str,
      delay_target: Union[int, bm.JaxArray, bm.ndarray]
  ):
    if name in self.local_delay_vars:
      return self.local_delay_vars[name].update(delay_target)
    else:
      if name not in self.global_delay_vars:
        raise ValueError(f'{name} is not defined in delay variables.')



# -*- coding: utf-8 -*-

from typing import Optional, Union, Sequence, Dict, List

from jax import vmap

import brainpy.math as bm
from brainpy.dyn.base import TwoEndConn, NeuGroup
from brainpy.initialize import Initializer, ZeroInit
from brainpy.tools.checking import check_sequence
from brainpy.types import Tensor

__all__ = [
  'DelayCoupling',
  'DiffusiveDelayCoupling',
  'AdditiveDelayCoupling',
]


class DelayCoupling(TwoEndConn):
  """
  Delay coupling base class.

  Parameters
  ----------
  coupling: str
    The way of coupling.
  gc: float
    The global coupling strength.
  signal_speed: float
    Signal transmission speed between areas.
  sc_mat: optional, tensor
    Structural connectivity matrix. Adjacency matrix of coupling strengths,
    will be normalized to 1. If not given, then a single node simulation
    will be assumed. Default None
  fl_mat: optional, tensor
    Fiber length matrix. Will be used for computing the
    delay matrix together with the signal transmission
    speed parameter `signal_speed`. Default None.

  """



  def __init__(
      self,
      pre: NeuGroup,
      post: NeuGroup,
      from_to: Union[str, Sequence[str]],
      conn_mat: Tensor,
      delay_mat: Optional[Tensor] = None,
      delay_initializer: Initializer = ZeroInit(),
      domain: str = 'local',
      name: str = None
  ):
    super(DelayCoupling, self).__init__(pre, post, name=name)

    # domain
    if domain not in ['global', 'local']:
      raise ValueError('"domain" must be a string in ["global", "local"]. '
                       f'Bug we got {domain}.')
    self.domain = domain

    # pairs of (source, destination)
    self.source_target_pairs: Dict[str, List[bm.Variable]] = dict()
    source_vars = {}
    if isinstance(from_to, str):
      from_to = [from_to]
    check_sequence(from_to, 'from_to', elem_type=str, allow_none=False)
    for pair in from_to:
      splits = [v.strip() for v in pair.split('->')]
      if len(splits) != 2:
        raise ValueError('The (source, target) pair in "from_to" '
                         'should be defined as "a -> b".')
      if not hasattr(self.pre, splits[0]):
        raise ValueError(f'"{splits[0]}" is not defined in pre-synaptic group {self.pre.name}')
      if not hasattr(self.post, splits[1]):
        raise ValueError(f'"{splits[1]}" is not defined in post-synaptic group {self.post.name}')
      source = f'{self.pre.name}.{splits[0]}'
      target = getattr(self.post, splits[1])
      if splits[0] not in self.source_target_pairs:
        self.source_target_pairs[source] = [target]
        source_vars[source] = getattr(self.pre, splits[0])
        if not isinstance(source_vars[source], bm.Variable):
          raise ValueError(f'The target variable {source} for delay should '
                           f'be an instance of brainpy.math.Variable, while '
                           f'we got {type(source_vars[source])}')
      else:
        if target in self.source_target_pairs:
          raise ValueError(f'{pair} has been defined twice in {from_to}.')
        self.source_target_pairs[source].append(target)

    # Connection matrix
    conn_mat = bm.asarray(conn_mat)
    required_shape = (self.post.num, self.pre.num)
    if conn_mat.shape != required_shape:
      raise ValueError(f'we expect the structural connection matrix has the shape of '
                       f'(post.num, pre.num), i.e., {required_shape}, '
                       f'while we got {conn_mat.shape}.')
    self.conn_mat = bm.asarray(conn_mat)
    bm.fill_diagonal(self.conn_mat, 0)

    # Delay matrix
    if delay_mat is None:
      self.delay_mat = bm.zeros(required_shape, dtype=bm.int_)
    else:
      if delay_mat.shape != required_shape:
        raise ValueError(f'we expect the fiber length matrix has the shape of '
                         f'(post.num, pre.num), i.e., {required_shape}. '
                         f'While we got {delay_mat.shape}.')
      self.delay_mat = bm.asarray(delay_mat, dtype=bm.int_)

    # delay variables
    num_delay_step = int(self.delay_mat.max())
    for var in self.source_target_pairs.keys():
      if domain == 'local':
        variable = source_vars[var]
        shape = (num_delay_step,) + variable.shape
        delay_data = delay_initializer(shape, dtype=variable.dtype)
        self.local_delay_vars[var] = bm.LengthDelay(variable, num_delay_step, delay_data)
      else:
        if var not in self.global_delay_vars:
          variable = source_vars[var]
          shape = (num_delay_step,) + variable.shape
          delay_data = delay_initializer(shape, dtype=variable.dtype)
          self.global_delay_vars[var] = bm.LengthDelay(variable, num_delay_step, delay_data)
          # save into local delay vars when first seen "var",
          # for later update current value!
          self.local_delay_vars[var] = self.global_delay_vars[var]
        else:
          if self.global_delay_vars[var].num_delay_step - 1 < num_delay_step:
            variable = source_vars[var]
            shape = (num_delay_step,) + variable.shape
            delay_data = delay_initializer(shape, dtype=variable.dtype)
            self.global_delay_vars[var].init(variable, num_delay_step, delay_data)

    self.register_implicit_nodes(self.local_delay_vars)
    self.register_implicit_nodes(self.global_delay_vars)

  def update(self, _t, _dt):
    raise NotImplementedError('Must implement the update() function by users.')


class DiffusiveDelayCoupling(DelayCoupling):
  def update(self, _t, _dt):
    for source, targets in self.source_target_pairs.items():
      # delay variable
      if self.domain == 'local':
        delay_var: bm.LengthDelay = self.local_delay_vars[source]
      elif self.domain == 'global':
        delay_var: bm.LengthDelay = self.global_delay_vars[source]
      else:
        raise ValueError(f'Unknown domain: {self.domain}')

      # current data
      name, var = source.split('.')
      assert name == self.pre.name
      variable = getattr(self.pre, var)

      # delays
      f = vmap(lambda i: delay_var(self.delay_mat[i], bm.arange(self.pre.num)))  # (pre.num,)
      delays = f(bm.arange(self.post.num).value)
      diffusive = delays - bm.expand_dims(variable, axis=1)  # (post.num, pre.num)
      diffusive = (self.conn_mat * diffusive).sum(axis=1)

      # output to target variable
      for target in targets:
        target.value += diffusive

      # update
      if source in self.local_delay_vars:
        delay_var.update(variable)


class AdditiveDelayCoupling(DelayCoupling):
  def update(self, _t, _dt):
    for source, targets in self.source_target_pairs.items():
      # delay variable
      if self.domain == 'local':
        delay_var: bm.LengthDelay = self.local_delay_vars[source]
      elif self.domain == 'global':
        delay_var: bm.LengthDelay = self.global_delay_vars[source]
      else:
        raise ValueError(f'Unknown domain: {self.domain}')

      # current data
      name, var = source.split('.')
      assert name == self.pre.name
      variable = getattr(self.pre, var)

      # delay function
      f = vmap(lambda i: delay_var(self.delay_mat[i], bm.arange(self.pre.num)))  # (pre.num,)
      delays = f(bm.arange(self.post.num))  # (post.num, pre.num)
      additive = (self.conn_mat * delays).sum(axis=1)

      # output to target variable
      for target in targets:
        target.value += additive

      # update
      if source in self.local_delay_vars:
        delay_var.update(variable)



# -*- coding: utf-8 -*-

from brainpy.integrators import constants, utils
from brainpy.integrators.sde.base import SDEIntegrator
from .generic import register_sde_integrator

__all__ = [
  'SRK1W1',
  'SRK2W1',
  'KlPl',
]


def _noise_terms(code_lines, variables, triple_integral=True):
  # num_vars = len(variables)
  # if num_vars > 1:
  #     code_lines.append(f'  all_I1 = math.normal(0.0, dt_sqrt, ({num_vars},)+math.shape({variables[0]}))')
  #     code_lines.append(f'  all_I0 = math.normal(0.0, dt_sqrt, ({num_vars},)+math.shape({variables[0]}))')
  #     code_lines.append(f'  all_I10 = 0.5 * {constants.DT} * (all_I1 + all_I0 / 3.0 ** 0.5)')
  #     code_lines.append(f'  all_I11 = 0.5 * (all_I1 ** 2 - {constants.DT})')
  #     if triple_integral:
  #         code_lines.append(f'  all_I111 = (all_I1 ** 3 - 3 * {constants.DT} * all_I1) / 6')
  #     code_lines.append(f'  ')
  #     for i, var in enumerate(variables):
  #         code_lines.append(f'  {var}_I1 = all_I1[{i}]')
  #         code_lines.append(f'  {var}_I0 = all_I0[{i}]')
  #         code_lines.append(f'  {var}_I10 = all_I10[{i}]')
  #         code_lines.append(f'  {var}_I11 = all_I11[{i}]')
  #         if triple_integral:
  #             code_lines.append(f'  {var}_I111 = all_I111[{i}]')
  #         code_lines.append(f'  ')
  # else:
  #     var = variables[0]
  #     code_lines.append(f'  {var}_I1 = math.normal(0.0, dt_sqrt, math.shape({var}))')
  #     code_lines.append(f'  {var}_I0 = math.normal(0.0, dt_sqrt, math.shape({var}))')
  #     code_lines.append(f'  {var}_I10 = 0.5 * {constants.DT} * ({var}_I1 + {var}_I0 / 3.0 ** 0.5)')
  #     code_lines.append(f'  {var}_I11 = 0.5 * ({var}_I1 ** 2 - {constants.DT})')
  #     if triple_integral:
  #         code_lines.append(f'  {var}_I111 = ({var}_I1 ** 3 - 3 * {constants.DT} * {var}_I1) / 6')
  #     code_lines.append('  ')

  for var in variables:
    code_lines.append(f'  {var}_I1 = random.normal(0.000, dt_sqrt, math.shape({var}))')
    code_lines.append(f'  {var}_I0 = random.normal(0.000, dt_sqrt, math.shape({var}))')
    code_lines.append(f'  {var}_I10 = 0.5 * {constants.DT} * ({var}_I1 + {var}_I0 / 3.0 ** 0.5)')
    code_lines.append(f'  {var}_I11 = 0.5 * ({var}_I1 ** 2 - {constants.DT})')
    if triple_integral:
      code_lines.append(f'  {var}_I111 = ({var}_I1 ** 3 - 3 * {constants.DT} * {var}_I1) / 6')
    code_lines.append('  ')


def _state1(code_lines, variables, parameters):
  f_names = [f'{var}_f_H0s1' for var in variables]
  g_names = [f'{var}_g_H1s1' for var in variables]
  code_lines.append(f'  {", ".join(f_names)} = f({", ".join(variables + parameters)})')
  code_lines.append(f'  {", ".join(g_names)} = g({", ".join(variables + parameters)})')
  code_lines.append('  ')


class SRK1W1(SDEIntegrator):
  r"""Order 2.0 weak SRK methods for SDEs with scalar Wiener process.

  This method has have strong orders :math:`(p_d, p_s) = (2.0,1.5)`.

  The Butcher table is:

  .. math::

      \begin{array}{l|llll|llll|llll}
          0   &&&&&  &&&&  &&&& \\
          3/4 &3/4&&&& 3/2&&& &&&& \\
          0   &0&0&0&& 0&0&0&& &&&&\\
          \hline
          0 \\
          1/4 & 1/4&&& & 1/2&&&\\
          1 & 1&0&&& -1&0&\\
          1/4& 0&0&1/4&&  -5&3&1/2\\
          \hline
          & 1/3& 2/3& 0 & 0 & -1 & 4/3 & 2/3&0 & -1 &4/3 &-1/3 &0 \\
          \hline
          & &&&& 2 &-4/3 & -2/3 & 0 & -2 & 5/3 & -2/3 & 1
      \end{array}


  References
  ----------

  .. [1] Rößler, Andreas. "Strong and weak approximation methods for stochastic differential
          equations—some recent developments." Recent developments in applied probability and
          statistics. Physica-Verlag HD, 2010. 127-153.
  .. [2] Rößler, Andreas. "Runge–Kutta methods for the strong approximation of solutions of
          stochastic differential equations." SIAM Journal on Numerical Analysis 48.3
          (2010): 922-952.

  """

  def __init__(self, f, g, dt=None, name=None, show_code=False,
               var_type=None, intg_type=None, wiener_type=None, state_delays=None):
    super(SRK1W1, self).__init__(f=f, g=g, dt=dt, show_code=show_code, name=name,
                                 var_type=var_type, intg_type=intg_type,
                                 wiener_type=wiener_type, state_delays=state_delays)
    assert self.wiener_type == constants.SCALAR_WIENER
    self.build()

  def build(self):
    # 2. code lines
    self.code_lines.append(f'  {constants.DT}_sqrt = {constants.DT} ** 0.5')

    # 2.1 noise
    _noise_terms(self.code_lines, self.variables, triple_integral=True)

    # 2.2 stage 1
    _state1(self.code_lines, self.variables, self.parameters)

    # 2.3 stage 2
    all_H0s2, all_H1s2 = [], []
    for var in self.variables:
      self.code_lines.append(f'  {var}_H0s2 = {var} + {constants.DT} * 0.75 * {var}_f_H0s1 + '
                             f'1.5 * {var}_g_H1s1 * {var}_I10 / {constants.DT}')
      all_H0s2.append(f'{var}_H0s2')
      self.code_lines.append(f'  {var}_H1s2 = {var} + {constants.DT} * 0.25 * {var}_f_H0s1 + '
                             f'dt_sqrt * 0.5 * {var}_g_H1s1')
      all_H1s2.append(f'{var}_H1s2')
    all_H0s2.append(f't + 0.75 * {constants.DT}')  # t
    all_H1s2.append(f't + 0.25 * {constants.DT}')  # t
    f_names = [f'{var}_f_H0s2' for var in self.variables]
    self.code_lines.append(f'  {", ".join(f_names)} = f({", ".join(all_H0s2 + self.parameters[1:])})')
    g_names = [f'{var}_g_H1s2' for var in self.variables]
    self.code_lines.append(f'  {", ".join(g_names)} = g({", ".join(all_H1s2 + self.parameters[1:])})')
    self.code_lines.append('  ')

    # 2.4 state 3
    all_H1s3 = []
    for var in self.variables:
      self.code_lines.append(f'  {var}_H1s3 = {var} + {constants.DT} * {var}_f_H0s1 - dt_sqrt * {var}_g_H1s1')
      all_H1s3.append(f'{var}_H1s3')
    all_H1s3.append(f't + {constants.DT}')  # t
    g_names = [f'{var}_g_H1s3' for var in self.variables]
    self.code_lines.append(f'  {", ".join(g_names)} = g({", ".join(all_H1s3 + self.parameters[1:])})')
    self.code_lines.append('  ')

    # 2.5 state 4
    all_H1s4 = []
    for var in self.variables:
      self.code_lines.append(f'  {var}_H1s4 = {var} + 0.25 * {constants.DT} * {var}_f_H0s1 + dt_sqrt * '
                             f'(-5 * {var}_g_H1s1 + 3 * {var}_g_H1s2 + 0.5 * {var}_g_H1s3)')
      all_H1s4.append(f'{var}_H1s4')
    all_H1s4.append(f't + 0.25 * {constants.DT}')  # t
    g_names = [f'{var}_g_H1s4' for var in self.variables]
    self.code_lines.append(f'  {", ".join(g_names)} = g({", ".join(all_H1s4 + self.parameters[1:])})')
    self.code_lines.append('  ')

    # 2.6 final stage
    for var in self.variables:
      self.code_lines.append(f'  {var}_f1 = {var}_f_H0s1/3 + {var}_f_H0s2 * 2/3')
      self.code_lines.append(
        f'  {var}_g1 = -{var}_I1 - {var}_I11/dt_sqrt + 2 * {var}_I10/{constants.DT} - 2 * {var}_I111/{constants.DT}')
      self.code_lines.append(f'  {var}_g2 = {var}_I1 * 4/3 + {var}_I11 / dt_sqrt * 4/3 - '
                             f'{var}_I10 / {constants.DT} * 4/3 + {var}_I111 / {constants.DT} * 5/3')
      self.code_lines.append(f'  {var}_g3 = {var}_I1 * 2/3 - {var}_I11/dt_sqrt/3 - '
                             f'{var}_I10 / {constants.DT} * 2/3 - {var}_I111 / {constants.DT} * 2/3')
      self.code_lines.append(f'  {var}_g4 = {var}_I111 / {constants.DT}')
      self.code_lines.append(f'  {var}_new = {var} + {constants.DT} * {var}_f1 + {var}_g1 * {var}_g_H1s1 + '
                             f'{var}_g2 * {var}_g_H1s2 + {var}_g3 * {var}_g_H1s3 + {var}_g4 * {var}_g_H1s4')
      self.code_lines.append('  ')

    # returns
    new_vars = [f'{var}_new' for var in self.variables]
    self.code_lines.append(f'  return {", ".join(new_vars)}')

    # return and compile
    self.integral = utils.compile_code(
      code_scope={k: v for k, v in self.code_scope.items()},
      code_lines=self.code_lines,
      show_code=self.show_code,
      func_name=self.func_name)


register_sde_integrator('srk1w1', SRK1W1)


class SRK2W1(SDEIntegrator):
  r"""Order 1.5 Strong SRK Methods for SDEs with Scalar Noise.

  This method has have strong orders :math:`(p_d, p_s) = (3.0,1.5)`.

  The Butcher table is:

  .. math::

      \begin{array}{c|cccc|cccc|ccc|}
          0 & 0 & 0 & 0 & 0 & 0 & 0 & 0 & 0 & & & & \\
          1 & 1 & 0 & 0 & 0 & 0 & 0 & 0 & 0 & & & & \\
          1 / 2 & 1 / 4 & 1 / 4 & 0 & 0 & 1 & 1 / 2 & 0 & 0 & & & & \\
          0 & 0 & 0 & 0 & 0 & 0 & 0 & 0 & 0 & & & & \\
          \hline 0 & 0 & 0 & 0 & 0 & 0 & 0 & 0 & 0 & & & & \\
          1 / 4 & 1 / 4 & 0 & 0 & 0 & -1 / 2 & 0 & 0 & 0 & & & & \\
          1 & 1 & 0 & 0 & 0 & 1 & 0 & 0 & 0 & & & & \\
          1 / 4 & 0 & 0 & 1 / 4 & 0 & 2 & -1 & 1 / 2 & 0 & & & & \\
          \hline & 1 / 6 & 1 / 6 & 2 / 3 & 0 & -1 & 4 / 3 & 2 / 3 & 0 & -1 & -4 / 3 & 1 / 3 & 0 \\
          \hline & & & & &2 & -4 / 3 & -2 / 3 & 0 & -2 & 5 / 3 & -2 / 3 & 1
      \end{array}


  References
  ----------

  .. [1] Rößler, Andreas. "Strong and weak approximation methods for stochastic differential
         equations—some recent developments." Recent developments in applied probability and
         statistics. Physica-Verlag HD, 2010. 127-153.
  .. [2] Rößler, Andreas. "Runge–Kutta methods for the strong approximation of solutions of
         stochastic differential equations." SIAM Journal on Numerical Analysis 48.3
         (2010): 922-952.
  """

  def __init__(self, f, g, dt=None, name=None, show_code=False,
               var_type=None, intg_type=None, wiener_type=None, state_delays=None):
    super(SRK2W1, self).__init__(f=f, g=g, dt=dt, show_code=show_code, name=name,
                                 var_type=var_type, intg_type=intg_type,
                                 wiener_type=wiener_type, state_delays=state_delays)
    assert self.wiener_type == constants.SCALAR_WIENER
    self.build()

  def build(self):
    self.code_lines.append(f'  {constants.DT}_sqrt = {constants.DT} ** 0.5')

    # 2.1 noise
    _noise_terms(self.code_lines, self.variables, triple_integral=True)

    # 2.2 stage 1
    _state1(self.code_lines, self.variables, self.parameters)

    # 2.3 stage 2
    # ----
    # H0s2 = x + dt * f_H0s1
    # H1s2 = x + dt * 0.25 * f_H0s1 - dt_sqrt * 0.5 * g_H1s1
    # f_H0s2 = f(H0s2, t + dt, *args)
    # g_H1s2 = g(H1s2, t + 0.25 * dt, *args)
    all_H0s2, all_H1s2 = [], []
    for var in self.variables:
      self.code_lines.append(f'  {var}_H0s2 = {var} + {constants.DT} * {var}_f_H0s1')
      all_H0s2.append(f'{var}_H0s2')
      self.code_lines.append(f'  {var}_H1s2 = {var} + {constants.DT} * 0.25 * {var}_f_H0s1 - '
                             f'dt_sqrt * 0.5 * {var}_g_H1s1')
      all_H1s2.append(f'{var}_H1s2')
    all_H0s2.append(f't + {constants.DT}')  # t
    all_H1s2.append(f't + 0.25 * {constants.DT}')  # t
    f_names = [f'{var}_f_H0s2' for var in self.variables]
    self.code_lines.append(f'  {", ".join(f_names)} = f({", ".join(all_H0s2 + self.parameters[1:])})')
    g_names = [f'{var}_g_H1s2' for var in self.variables]
    self.code_lines.append(f'  {", ".join(g_names)} = g({", ".join(all_H1s2 + self.parameters[1:])})')
    self.code_lines.append('  ')

    # 2.4 state 3
    # ---
    # H0s3 = x + dt * (0.25 * f_H0s1 + 0.25 * f_H0s2) + (g_H1s1 + 0.5 * g_H1s2) * I10 / dt
    # H1s3 = x + dt * f_H0s1 + dt_sqrt * g_H1s1
    # f_H0s3 = g(H0s3, t + 0.5 * dt, *args)
    # g_H1s3 = g(H1s3, t + dt, *args)
    all_H0s3, all_H1s3 = [], []
    for var in self.variables:
      self.code_lines.append(f'  {var}_H0s3 = {var} + {constants.DT} * (0.25 * {var}_f_H0s1 + 0.25 * {var}_f_H0s2) + '
                             f'({var}_g_H1s1 + 0.5 * {var}_g_H1s2) * {var}_I10 / {constants.DT}')
      all_H0s3.append(f'{var}_H0s3')
      self.code_lines.append(f'  {var}_H1s3 = {var} + {constants.DT} * {var}_f_H0s1 + dt_sqrt * {var}_g_H1s1')
      all_H1s3.append(f'{var}_H1s3')
    all_H0s3.append(f't + 0.5 * {constants.DT}')  # t
    all_H1s3.append(f't + {constants.DT}')  # t
    f_names = [f'{var}_f_H0s3' for var in self.variables]
    g_names = [f'{var}_g_H1s3' for var in self.variables]
    self.code_lines.append(f'  {", ".join(f_names)} = f({", ".join(all_H0s3 + self.parameters[1:])})')
    self.code_lines.append(f'  {", ".join(g_names)} = g({", ".join(all_H1s3 + self.parameters[1:])})')
    self.code_lines.append('  ')

    # 2.5 state 4
    # ----
    # H1s4 = x + dt * 0.25 * f_H0s3 + dt_sqrt * (2 * g_H1s1 - g_H1s2 + 0.5 * g_H1s3)
    # g_H1s4 = g(H1s4, t + 0.25 * dt, *args)
    all_H1s4 = []
    for var in self.variables:
      self.code_lines.append(f'  {var}_H1s4 = {var} + 0.25 * {constants.DT} * {var}_f_H0s1 + dt_sqrt * '
                             f'(2 * {var}_g_H1s1 - {var}_g_H1s2 + 0.5 * {var}_g_H1s3)')
      all_H1s4.append(f'{var}_H1s4')
    all_H1s4.append(f't + 0.25 * {constants.DT}')  # t
    g_names = [f'{var}_g_H1s4' for var in self.variables]
    self.code_lines.append(f'  {", ".join(g_names)} = g({", ".join(all_H1s4 + self.parameters[1:])})')
    self.code_lines.append('  ')

    # 2.6 final stage
    # ----
    # f1 = f_H0s1 / 6 + f_H0s2 / 6 + f_H0s3 * 2 / 3
    # g1 = - I1 + I11 / dt_sqrt + 2 * I10 / dt - 2 * I111 / dt
    # g2 = I1 * 4 / 3 - I11 / dt_sqrt * 4 / 3 - I10 / dt * 4 / 3 + I111 / dt * 5 / 3
    # g3 = I1 * 2 / 3 + I11 / dt_sqrt / 3 - I10 / dt * 2 / 3 - I111 / dt * 2 / 3
    # g4 = I111 / dt
    # y1 = x + dt * f1 + g1 * g_H1s1 + g2 * g_H1s2 + g3 * g_H1s3 + g4 * g_H1s4
    for var in self.variables:
      self.code_lines.append(f'  {var}_f1 = {var}_f_H0s1/6 + {var}_f_H0s2/6 + {var}_f_H0s3*2/3')
      self.code_lines.append(
        f'  {var}_g1 = -{var}_I1 + {var}_I11/dt_sqrt + 2 * {var}_I10/{constants.DT} - 2 * {var}_I111/{constants.DT}')
      self.code_lines.append(f'  {var}_g2 = {var}_I1 * 4/3 - {var}_I11 / dt_sqrt * 4/3 - '
                             f'{var}_I10 / {constants.DT} * 4/3 + {var}_I111 / {constants.DT} * 5/3')
      self.code_lines.append(f'  {var}_g3 = {var}_I1 * 2/3 + {var}_I11/dt_sqrt/3 - '
                             f'{var}_I10 / {constants.DT} * 2/3 - {var}_I111 / {constants.DT} * 2/3')
      self.code_lines.append(f'  {var}_g4 = {var}_I111 / {constants.DT}')
      self.code_lines.append(f'  {var}_new = {var} + {constants.DT} * {var}_f1 + {var}_g1 * {var}_g_H1s1 + '
                             f'{var}_g2 * {var}_g_H1s2 + {var}_g3 * {var}_g_H1s3 + {var}_g4 * {var}_g_H1s4')
      self.code_lines.append('  ')

    # returns
    new_vars = [f'{var}_new' for var in self.variables]
    self.code_lines.append(f'  return {", ".join(new_vars)}')

    # return and compile
    self.integral = utils.compile_code(
      code_scope={k: v for k, v in self.code_scope.items()},
      code_lines=self.code_lines,
      show_code=self.show_code,
      func_name=self.func_name)


register_sde_integrator('srk2w1', SRK2W1)


class KlPl(SDEIntegrator):
  def __init__(self, f, g, dt=None, name=None, show_code=False,
               var_type=None, intg_type=None, wiener_type=None, state_delays=None):
    super(KlPl, self).__init__(f=f, g=g, dt=dt, show_code=show_code, name=name,
                               var_type=var_type, intg_type=intg_type,
                               wiener_type=wiener_type, state_delays=state_delays)
    assert self.wiener_type == constants.SCALAR_WIENER
    self.build()

  def build(self):
    self.code_lines.append(f'  {constants.DT}_sqrt = {constants.DT} ** 0.5')

    # 2.1 noise
    _noise_terms(self.code_lines, self.variables, triple_integral=False)

    # 2.2 stage 1
    _state1(self.code_lines, self.variables, self.parameters)

    # 2.3 stage 2
    # ----
    # H1s2 = x + dt * f_H0s1 + dt_sqrt * g_H1s1
    # g_H1s2 = g(H1s2, t0, *args)
    all_H1s2 = []
    for var in self.variables:
      self.code_lines.append(f'  {var}_H1s2 = {var} + {constants.DT} * {var}_f_H0s1 + dt_sqrt * {var}_g_H1s1')
      all_H1s2.append(f'{var}_H1s2')
    g_names = [f'{var}_g_H1s2' for var in self.variables]
    self.code_lines.append(f'  {", ".join(g_names)} = g({", ".join(all_H1s2 + self.parameters)})')
    self.code_lines.append('  ')

    # 2.4 final stage
    # ----
    # g1 = (I1 - I11 / dt_sqrt + I10 / dt)
    # g2 = I11 / dt_sqrt
    # y1 = x + dt * f_H0s1 + g1 * g_H1s1 + g2 * g_H1s2
    for var in self.variables:
      self.code_lines.append(f'  {var}_g1 = -{var}_I1 + {var}_I11/dt_sqrt + {var}_I10/{constants.DT}')
      self.code_lines.append(f'  {var}_g2 = {var}_I11 / dt_sqrt')
      self.code_lines.append(f'  {var}_new = {var} + {constants.DT} * {var}_f_H0s1 + '
                             f'{var}_g1 * {var}_g_H1s1 + {var}_g2 * {var}_g_H1s2')
      self.code_lines.append('  ')

    # returns
    new_vars = [f'{var}_new' for var in self.variables]
    self.code_lines.append(f'  return {", ".join(new_vars)}')

    # return and compile
    self.integral = utils.compile_code(
      code_scope={k: v for k, v in self.code_scope.items()},
      code_lines=self.code_lines,
      show_code=self.show_code,
      func_name=self.func_name)


register_sde_integrator('klpl', KlPl)



# -*- coding: utf-8 -*-

from functools import partial

import numpy as np
from jax import vmap, jit, lax, numpy as jnp

from brainpy import math as bm

__all__ = [
  'cross_correlation',
  'voltage_fluctuation',
  'matrix_correlation',
  'weighted_correlation',
  'functional_connectivity',
  'functional_connectivity_dynamics',
]


@jit
@partial(vmap, in_axes=(None, 0, 0))
def _cc(states, i, j):
  sqrt_ij = jnp.sqrt(jnp.sum(states[i]) * jnp.sum(states[j]))
  return lax.cond(sqrt_ij == 0.,
                  lambda _: 0.,
                  lambda ij: jnp.sum(states[i] * states[j]) / sqrt_ij,
                  (i, j))


def cross_correlation(spikes, bin, dt=None):
  r"""Calculate cross correlation index between neurons.

  The coherence [1]_ between two neurons i and j is measured by their
  cross-correlation of spike trains at zero time lag within a time bin
  of :math:`\Delta t = \tau`. More specifically, suppose that a long
  time interval T is divided into small bins of :math:`\Delta t` and
  that two spike trains are given by :math:`X(l)=` 0 or 1, :math:`Y(l)=` 0
  or 1, :math:`l=1,2, \ldots, K(T / K=\tau)`. Thus, we define a coherence
  measure for the pair as:

  .. math::

      \kappa_{i j}(\tau)=\frac{\sum_{l=1}^{K} X(l) Y(l)}
      {\sqrt{\sum_{l=1}^{K} X(l) \sum_{l=1}^{K} Y(l)}}

  The population coherence measure :math:`\kappa(\tau)` is defined by the
  average of :math:`\kappa_{i j}(\tau)` over many pairs of neurons in the
  network.

  Parameters
  ----------
  spikes :
      The history of spike states of the neuron group.
      It can be easily get via `StateMonitor(neu, ['spike'])`.
  bin : float, int
      The time bin to normalize spike states.
  dt : float, optional
      The time precision.

  Returns
  -------
  cc_index : float
      The cross correlation value which represents the synchronization index.

  References
  ----------
  .. [1] Wang, Xiao-Jing, and György Buzsáki. "Gamma oscillation by synaptic
         inhibition in a hippocampal interneuronal network model." Journal of
         neuroscience 16.20 (1996): 6402-6413.
  """
  spikes = bm.as_device_array(spikes)
  dt = bm.get_dt() if dt is None else dt
  bin_size = int(bin / dt)
  num_hist, num_neu = spikes.shape
  num_bin = int(np.ceil(num_hist / bin_size))
  if num_bin * bin_size != num_hist:
    spikes = jnp.append(spikes, jnp.zeros((num_bin * bin_size - num_hist, num_neu)), axis=0)
  states = spikes.T.reshape((num_neu, num_bin, bin_size))
  states = jnp.asarray(jnp.sum(states, axis=2) > 0., dtype=jnp.float_)
  indices = jnp.tril_indices(4, k=-1)
  return jnp.mean(_cc(states, *indices))


@partial(vmap, in_axes=(None, 0))
def _var(neu_signal, i):
  neu_signal = neu_signal[:, i]
  return jnp.mean(neu_signal * neu_signal) - jnp.mean(neu_signal) ** 2


@jit
def voltage_fluctuation(potentials):
  r"""Calculate neuronal synchronization via voltage variance.

  The method comes from [1]_ [2]_ [3]_.

  First, average over the membrane potential :math:`V`

  .. math::

      V(t) = \frac{1}{N} \sum_{i=1}^{N} V_i(t)

  The variance of the time fluctuations of :math:`V(t)` is

  .. math::

      \sigma_V^2 = \left\langle \left[ V(t) \right]^2 \right\rangle_t -
      \left[ \left\langle V(t) \right\rangle_t \right]^2

  where :math:`\left\langle \ldots \right\rangle_t = (1 / T_m) \int_0^{T_m} dt \, \ldots`
  denotes time-averaging over a large time, :math:`\tau_m`. After normalization
  of :math:`\sigma_V` to the average over the population of the single cell
  membrane potentials

  .. math::

      \sigma_{V_i}^2 = \left\langle\left[ V_i(t) \right]^2 \right\rangle_t -
      \left[ \left\langle V_i(t) \right\rangle_t \right]^2

  one defines a synchrony measure, :math:`\chi (N)`, for the activity of a system
  of :math:`N` neurons by:

  .. math::

      \chi^2 \left( N \right) = \frac{\sigma_V^2}{ \frac{1}{N} \sum_{i=1}^N
      \sigma_{V_i}^2}

  Parameters
  ----------
  potentials :
      The membrane potential matrix of the neuron group.

  Returns
  -------
  sync_index : float
      The synchronization index.

  References
  ----------
  .. [1] Golomb, D. and Rinzel J. (1993) Dynamics of globally coupled
         inhibitory neurons with heterogeneity. Phys. Rev. reversal_potential 48:4810-4814.
  .. [2] Golomb D. and Rinzel J. (1994) Clustering in globally coupled
         inhibitory neurons. Physica D 72:259-282.
  .. [3] David Golomb (2007) Neuronal synchrony measures. Scholarpedia, 2(1):1347.
  """

  potentials = bm.as_device_array(potentials)
  num_hist, num_neu = potentials.shape
  var_mean = jnp.mean(_var(potentials, jnp.arange(num_neu)))
  avg = jnp.mean(potentials, axis=1)
  avg_var = jnp.mean(avg * avg) - jnp.mean(avg) ** 2
  return lax.cond(var_mean != 0.,
                  lambda _: avg_var / var_mean,
                  lambda _: 1.,
                  ())


def matrix_correlation(x, y):
  """Pearson correlation of the lower triagonal of two matrices.

    The triangular matrix is offset by k = 1 in order to ignore the diagonal line

  Parameters
  ----------
  x: tensor
    First matrix.
  y: tensor
    Second matrix

  Returns
  -------
  coef: tensor
    Correlation coefficient
  """
  x = bm.as_numpy(x)
  y = bm.as_numpy(y)
  if x.ndim != 2:
    raise ValueError(f'Only support 2d tensor, but we got a tensor '
                     f'with the shape of {x.shape}')
  if y.ndim != 2:
    raise ValueError(f'Only support 2d tensor, but we got a tensor '
                     f'with the shape of {y.shape}')
  x = x[np.triu_indices_from(x, k=1)]
  y = y[np.triu_indices_from(y, k=1)]
  cc = np.corrcoef(x, y)[0, 1]
  return cc


def functional_connectivity(activities):
  """Functional connectivity matrix of timeseries activities.

  Parameters
  ----------
  activities: tensor
    The multidimensional tensor with the shape of ``(num_time, num_sample)``.

  Returns
  -------
  connectivity_matrix: tensor
    ``num_sample x num_sample`` functional connectivity matrix.
  """
  activities = bm.as_numpy(activities)
  if activities.ndim != 2:
    raise ValueError('Only support 2d tensor with shape of "(num_time, num_sample)". '
                     f'But we got a tensor with the shape of {activities.shape}')
  fc = np.corrcoef(activities.T)
  return np.nan_to_num(fc)


@jit
def functional_connectivity_dynamics(activities, window_size=30, step_size=5):
  """Computes functional connectivity dynamics (FCD) matrix.

  Parameters
  ----------
  activities: tensor
    The time series with shape of ``(num_time, num_sample)``.
  window_size: int
    Size of each rolling window in time steps, defaults to 30.
  step_size: int
    Step size between each rolling window, defaults to 5.

  Returns
  -------
  fcd_matrix: tensor
    FCD matrix.
  """
  pass


def _weighted_mean(x, w):
  """Weighted Mean"""
  return jnp.sum(x * w) / jnp.sum(w)


def _weighted_cov(x, y, w):
  """Weighted Covariance"""
  return jnp.sum(w * (x - _weighted_mean(x, w)) * (y - _weighted_mean(y, w))) / jnp.sum(w)


@jit
def weighted_correlation(x, y, w):
  """Weighted Pearson correlation of two data series.

  Parameters
  ----------
  x: tensor
    The data series 1.
  y: tensor
    The data series 2.
  w: tensor
    Weight vector, must have same length as x and y.

  Returns
  -------
  corr: tensor
    Weighted correlation coefficient.
  """
  x = bm.as_device_array(x)
  y = bm.as_device_array(y)
  w = bm.as_device_array(w)
  if x.ndim != 1:
    raise ValueError(f'Only support 1d tensor, but we got a tensor '
                     f'with the shape of {x.shape}')
  if y.ndim != 1:
    raise ValueError(f'Only support 1d tensor, but we got a tensor '
                     f'with the shape of {y.shape}')
  if w.ndim != 1:
    raise ValueError(f'Only support 1d tensor, but we got a tensor '
                     f'with the shape of {w.shape}')
  return _weighted_cov(x, y, w) / jnp.sqrt(_weighted_cov(x, x, w) * _weighted_cov(y, y, w))



# -*- coding: utf-8 -*-

import jax.numpy as jnp
import numpy as np
from brainpy.tools.others import numba_jit


__all__ = [
  'find_indexes_of_limit_cycle_max',
  'euclidean_distance',
]


@numba_jit
def _f1(arr, grad, tol):
  condition = np.logical_and(grad[:-1] * grad[1:] <= 0, grad[:-1] >= 0)
  indexes = np.where(condition)[0]
  if len(indexes) >= 2:
    data = arr[indexes[-2]: indexes[-1]]
    length = np.max(data) - np.min(data)
    a = arr[indexes[-2]]
    b = arr[indexes[-1]]
    # TODO: how to choose length threshold, 1e-3?
    if length > 1e-3 and np.abs(a - b) <= tol * length:
      return indexes[-2:]
  return np.array([-1, -1])


def find_indexes_of_limit_cycle_max(arr, tol=0.001):
  grad = np.gradient(arr)
  return _f1(arr, grad, tol)


# @tools.numba_jit
def euclidean_distance(points: np.ndarray):
  """Get the distance matrix.

  Equivalent to:

  >>> from scipy.spatial.distance import squareform, pdist
  >>> f = lambda points: squareform(pdist(points, metric="euclidean"))

  Parameters
  ----------
  points: jnp.ndarray, bm.JaxArray
    The points.

  Returns
  -------
  dist_matrix: jnp.ndarray
    The distance matrix.
  """
  num_point = points.shape[0]
  indices = np.triu_indices(num_point)
  dist_mat = np.zeros((num_point, num_point))
  for idx in range(len(indices[0])):
    i = indices[0][idx]
    j = indices[1][idx]
    dist_mat[i, j] = np.linalg.norm(points[i] - points[j])
  dist_mat = np.maximum(dist_mat, dist_mat.T)
  return dist_mat




# -*- coding: utf-8 -*-


import jax.numpy as jnp

import brainpy.math as bm
from brainpy import errors
from brainpy.dyn.base import DynamicalSystem
from brainpy.dyn.runners import DSRunner
from brainpy.integrators.joint_eq import JointEq
from brainpy.integrators.ode.base import ODEIntegrator

__all__ = [
  'model_transform',
  'NumDSWrapper',
  'TrajectModel',
]


def model_transform(model):
  # check integrals
  if isinstance(model, NumDSWrapper):
    return model
  elif isinstance(model, ODEIntegrator):  #
    model = [model]

  # check model types
  if isinstance(model, (list, tuple)):
    if len(model) == 0:
      raise errors.AnalyzerError(f'Found no integrators: {model}')
    model = tuple(model)
    for intg in model:
      if not isinstance(intg, ODEIntegrator):
        raise errors.AnalyzerError(f'Must be the instance of {ODEIntegrator}, but got {intg}.')
  elif isinstance(model, dict):
    if len(model) == 0:
      raise errors.AnalyzerError(f'Found no integrators: {model}')
    model = tuple(model.values())
    for intg in model:
      if not isinstance(intg, ODEIntegrator):
        raise errors.AnalyzerError(f'Must be the instance of {ODEIntegrator}, but got {intg}')
  elif isinstance(model, DynamicalSystem):
    model = tuple(model.ints().subset(ODEIntegrator).unique().values())
  else:
    raise errors.UnsupportedError(f'Dynamics analysis by symbolic approach only supports '
                                  f'list/tuple/dict of {ODEIntegrator} or {DynamicalSystem}, '
                                  f'but we got: {type(model)}: {str(model)}')

  new_model = []
  for intg in model:
    if isinstance(intg.f, JointEq):
      new_model.extend([type(intg)(eq, var_type=intg.var_type, dt=intg.dt) for eq in intg.f.eqs])
    else:
      new_model.append(intg)

  # pars to update
  pars_update = set()
  for intg in new_model:
    pars_update.update(intg.parameters[1:])

  all_variables = set()
  all_parameters = set()
  for integral in new_model:
    if len(integral.variables) != 1:
      raise errors.AnalyzerError(f'Only supports one {ODEIntegrator.__name__} one variable, '
                                 f'but we got {len(integral.variables)} variables in {integral}.')
    var = integral.variables[0]
    if var in all_variables:
      raise errors.AnalyzerError(f'Variable name {var} has been defined before. '
                                 f'Please change another name.')
    all_variables.add(var)
    # parameters
    all_parameters.update(integral.parameters[1:])

  # form a dynamic model
  return NumDSWrapper(integrals=new_model,
                      variables=list(all_variables),
                      parameters=list(all_parameters),
                      pars_update=pars_update)


class NumDSWrapper(object):
  """The wrapper of a dynamical model."""

  def __init__(self,
               integrals,
               variables,
               parameters,
               pars_update=None):
    self.INTG = integrals  # all integrators
    self.F = {intg.variables[0]: intg.f for intg in integrals}  # all integrators
    self.variables = variables  # all variables
    self.parameters = parameters  # all parameters
    self.pars_update = pars_update  # the parameters to update
    self.name2integral = {intg.variables[0]: intg for intg in integrals}
    self.name2derivative = {intg.variables[0]: intg.f for intg in integrals}


class TrajectModel(DynamicalSystem):
  def __init__(self, integrals: dict, initial_vars: dict, pars=None, dt=None):
    super(TrajectModel, self).__init__()

    # variables
    assert isinstance(initial_vars, dict)
    initial_vars = {k: bm.Variable(jnp.asarray(bm.as_device_array(v), dtype=jnp.float_))
                    for k, v in initial_vars.items()}
    self.register_implicit_vars(initial_vars)

    # parameters
    pars = dict() if pars is None else pars
    assert isinstance(pars, dict)
    self.pars = [jnp.asarray(bm.as_device_array(v), dtype=jnp.float_)
                 for k, v in pars.items()]

    # integrals
    self.integrals = integrals

    # runner
    self.runner = DSRunner(self,
                           monitors=list(initial_vars.keys()),
                           dyn_vars=self.vars().unique(), dt=dt,
                           progress_bar=False)

  def update(self, _t, _dt):
    all_vars = list(self.implicit_vars.values())
    for key, intg in self.integrals.items():
      self.implicit_vars[key].update(intg(*all_vars, *self.pars, dt=_dt))

  def __getattr__(self, item):
    child_vars = super(TrajectModel, self).__getattribute__('implicit_vars')
    if item in child_vars:
      return child_vars[item]
    else:
      return super(TrajectModel, self).__getattribute__(item)

  def run(self, duration):
    self.runner.run(duration)
    return self.runner.mon



# -*- coding: utf-8 -*-

"""
This module provides methods to initialize weights.
You can access them through ``brainpy.init.XXX``.
"""

from .base import *
from .generic import *
from .random_inits import *
from .regular_inits import *
from .decay_inits import *



# -*- coding: utf-8 -*-

import abc

__all__ = [
  'Initializer',
  'InterLayerInitializer',
  'IntraLayerInitializer'
]


class Initializer(abc.ABC):
  """Base Initialization Class."""

  @abc.abstractmethod
  def __call__(self, shape, dtype=None):
    raise NotImplementedError

  def __repr__(self):
    return self.__class__.__name__


class InterLayerInitializer(Initializer):
  """The superclass of Initializers that initialize the weights between two layers."""

  @abc.abstractmethod
  def __call__(self, shape, dtype=None):
    raise NotImplementedError

  def __repr__(self):
    return self.__class__.__name__


class IntraLayerInitializer(Initializer):
  """The superclass of Initializers that initialize the weights within a layer."""

  @abc.abstractmethod
  def __call__(self, shape, dtype=None):
    raise NotImplementedError

  def __repr__(self):
    return self.__class__.__name__



# -*- coding: utf-8 -*-

import warnings

from brainpy.integrators import ode, sde

__all__ = [
  'set_default_odeint',
  'set_default_sdeint',
  'get_default_odeint',
  'get_default_sdeint',
]


def set_default_odeint(method):
  """Set default ode integrator.

  .. deprecated:: 2.1.0
     Please use "brainpy.ode.set_default_odeint" instead.
  """
  warnings.warn('Please use "brainpy.ode.set_default_odeint" instead. '
                '"brainpy.set_default_odeint" is deprecated since '
                'version 2.1.0', DeprecationWarning)
  ode.set_default_odeint(method)


def get_default_odeint():
  """Get default ode integrator.

  .. deprecated:: 2.1.0
     Please use "brainpy.ode.get_default_odeint" instead.
  """
  warnings.warn('Please use "brainpy.ode.get_default_odeint" instead. '
                '"brainpy.get_default_odeint" is deprecated since '
                'version 2.1.0', DeprecationWarning)
  ode.get_default_odeint()


def set_default_sdeint(method):
  """Set default sde integrator.

  .. deprecated:: 2.1.0
     Please use "brainpy.ode.set_default_sdeint" instead.
  """
  warnings.warn('Please use "brainpy.sde.set_default_sdeint" instead. '
                '"brainpy.set_default_sdeint" is deprecated since '
                'version 2.1.0', DeprecationWarning)
  sde.set_default_sdeint(method)


def get_default_sdeint():
  """Get default sde integrator.

  .. deprecated:: 2.1.0
     Please use "brainpy.ode.get_default_sdeint" instead.
  """
  warnings.warn('Please use "brainpy.sde.get_default_sdeint" instead. '
                '"brainpy.get_default_sdeint" is deprecated since '
                'version 2.1.0', DeprecationWarning)
  sde.get_default_sdeint()



# -*- coding: utf-8 -*-

import ast
from collections import Counter

from brainpy import errors, tools

import sympy
import sympy.functions.elementary.complexes
import sympy.functions.elementary.exponential
import sympy.functions.elementary.hyperbolic
import sympy.functions.elementary.integers
import sympy.functions.elementary.miscellaneous
import sympy.functions.elementary.trigonometric
from sympy.codegen import cfunctions
from sympy.printing.precedence import precedence
from sympy.printing.str import StrPrinter


__all__ = [
  'CONSTANT_NOISE',
  'FUNCTIONAL_NOISE',
  'FUNCTION_MAPPING',

  'Parser',
  'Printer',
  'str2sympy',
  'sympy2str',

  'Expression',
  'SingleDiffEq',
]


CONSTANT_NOISE = 'CONSTANT'
FUNCTIONAL_NOISE = 'FUNCTIONAL'

FUNCTION_MAPPING = {
  # functions in inherit python
  # ---------------------------
  'abs': sympy.functions.elementary.complexes.Abs,

  # functions in numpy
  # ------------------
  'sign': sympy.sign,
  'sinc': sympy.functions.elementary.trigonometric.sinc,
  'arcsin': sympy.functions.elementary.trigonometric.asin,
  'arccos': sympy.functions.elementary.trigonometric.acos,
  'arctan': sympy.functions.elementary.trigonometric.atan,
  'arctan2': sympy.functions.elementary.trigonometric.atan2,
  'arcsinh': sympy.functions.elementary.hyperbolic.asinh,
  'arccosh': sympy.functions.elementary.hyperbolic.acosh,
  'arctanh': sympy.functions.elementary.hyperbolic.atanh,

  'log2': cfunctions.log2,
  'log1p': cfunctions.log1p,

  'expm1': cfunctions.expm1,
  'exp2': cfunctions.exp2,

  # functions in backend
  # ------------------
  'asin': sympy.functions.elementary.trigonometric.asin,
  'acos': sympy.functions.elementary.trigonometric.acos,
  'atan': sympy.functions.elementary.trigonometric.atan,
  'atan2': sympy.functions.elementary.trigonometric.atan2,
  'asinh': sympy.functions.elementary.hyperbolic.asinh,
  'acosh': sympy.functions.elementary.hyperbolic.acosh,
  'atanh': sympy.functions.elementary.hyperbolic.atanh,

  # functions in both numpy and backend
  # --------------------------------

  'cos': sympy.functions.elementary.trigonometric.cos,
  'sin': sympy.functions.elementary.trigonometric.sin,
  'tan': sympy.functions.elementary.trigonometric.tan,
  'cosh': sympy.functions.elementary.hyperbolic.cosh,
  'sinh': sympy.functions.elementary.hyperbolic.sinh,
  'tanh': sympy.functions.elementary.hyperbolic.tanh,

  'log': sympy.functions.elementary.exponential.log,
  'log10': cfunctions.log10,
  'sqrt': sympy.functions.elementary.miscellaneous.sqrt,
  'exp': sympy.functions.elementary.exponential.exp,
  'hypot': cfunctions.hypot,

  'ceil': sympy.functions.elementary.integers.ceiling,
  'floor': sympy.functions.elementary.integers.floor,
}

CONSTANT_MAPPING = {
  # constants in both numpy and backend
  # --------------------------------
  'pi': sympy.pi,
  'e': sympy.E,
  'inf': sympy.S.Infinity,
}

_func2expr = {
  'int_': 'int',
  'sign': 'math.sign',
  'cos': 'math.cos',
  'sin': 'math.sin',
  'tan': 'math.tan',
  'sinc': 'math.sinc',
  'arcsin': 'math.arcsin',
  'arccos': 'math.arccos',
  'arctan': 'math.arctan',
  'arctan2': 'math.arctan2',
  'cosh': 'math.cosh',
  'sinh': 'math.cosh',
  'tanh': 'math.tanh',
  'arcsinh': 'math.arcsinh',
  'arccosh': 'math.arccosh',
  'arctanh': 'math.arctanh',
  'ceil': 'math.ceil',
  'floor': 'math.floor',
  'log': 'math.log',
  'log2': 'math.log2',
  'log1p': 'math.log1p',
  'log10': 'math.log10',
  'exp': 'math.exp',
  'expm1': 'math.expm1',
  'exp2': 'math.exp2',
  'hypot': 'math.hypot',
  'sqrt': 'math.sqrt',
}

_constants = {
  'pi': 'math.pi',
  'e': 'math.e',
  'inf': 'math.inf',
}


class Parser(object):
  expression_ops = {
    'Add': sympy.Add,
    'Mult': sympy.Mul,
    'Pow': sympy.Pow,
    'Mod': sympy.Mod,
    # Compare
    'Lt': sympy.StrictLessThan,
    'LtE': sympy.LessThan,
    'Gt': sympy.StrictGreaterThan,
    'GtE': sympy.GreaterThan,
    'Eq': sympy.Eq,
    'NotEq': sympy.Ne,
    # Bool math
    'And': sympy.And,
    'Or': sympy.Or,
    # BinOp
    'Sub': '-',
    'Div': '/',
    'FloorDiv': '//',
    # Compare
    'Not': 'not',
    'UAdd': '+',
    'USub': '-',
    # Augmented assign
    'AugAdd': '+=',
    'AugSub': '-=',
    'AugMult': '*=',
    'AugDiv': '/=',
    'AugPow': '**=',
    'AugMod': '%=',
  }

  def __init__(self, expr):
    self.contain_unknown_func = False
    node = ast.parse(expr.strip(), mode='eval')
    self.expr = self.render_node(node.body)

  def render_node(self, node):
    nodename = node.__class__.__name__
    methname = 'render_' + nodename
    try:
      return getattr(self, methname)(node)
    except AttributeError:
      raise SyntaxError(f"Unknown syntax: {nodename}, {node}")

  def render_Attribute(self, node):
    if node.attr in CONSTANT_MAPPING:
      return CONSTANT_MAPPING[node.attr]
    else:
      names = self._get_attr_value(node, [])
      return sympy.Symbol('.'.join(names), real=True)

  def render_Constant(self, node):
    if node.value is True or node.value is False or node.value is None:
      return self.render_NameConstant(node)
    else:
      return self.render_Num(node)

  def render_element_parentheses(self, node):
    """Render an element with parentheses around it or leave them away for
    numbers, names and function calls.
    """
    if node.__class__.__name__ in ['Name', 'NameConstant']:
      return self.render_node(node)
    elif node.__class__.__name__ in ['Num', 'Constant'] and \
        getattr(node, 'n', getattr(node, 'value', None)) >= 0:
      return self.render_node(node)
    elif node.__class__.__name__ == 'Call':
      return self.render_node(node)
    else:
      return f'({self.render_node(node)})'

  def render_BinOp_parentheses(self, left, right, op):
    """Use a simplified checking whether it is possible to omit parentheses:
    only omit parentheses for numbers, variable names or function calls.
    This means we still put needless parentheses because we ignore
    precedence rules, e.g. we write "3 + (4 * 5)" but at least we do
    not do "(3) + ((4) + (5))" """
    ops = {'BitXor': ('^', '**'),
           'BitAnd': ('&', 'and'),
           'BitOr': ('|', 'or')}
    op_class = op.__class__.__name__
    # Give a more useful error message when using bit-wise operators
    if op_class in ['BitXor', 'BitAnd', 'BitOr']:
      correction = ops.get(op_class)
      raise SyntaxError('The operator "{}" is not supported, use "{}" '
                        'instead.'.format(correction[0], correction[1]))
    return f'{self.render_element_parentheses(left)} ' \
           f'{self.expression_ops[op_class]} ' \
           f'{self.render_element_parentheses(right)}'

  def render_Assign(self, node):
    if len(node.targets) > 1:
      raise SyntaxError("Only support syntax like a=b not a=b=c")
    return f'{self.render_node(node.targets[0])} = {self.render_node(node.value)}'

  def render_AugAssign(self, node):
    target = node.target.id
    rhs = self.render_node(node.value)
    op = self.expression_ops['Aug' + node.op.__class__.__name__]
    return f'{target} {op} {rhs}'

  def _get_attr_value(self, node, names):
    if hasattr(node, 'value'):
      names.insert(0, node.attr)
      return self._get_attr_value(node.value, names)
    else:
      assert hasattr(node, 'id')
      names.insert(0, node.id)
      return names

  def render_func(self, node):
    if hasattr(node, 'id'):
      if node.id in FUNCTION_MAPPING:
        f = FUNCTION_MAPPING[node.id]
        return f
      # special workaround for the "int" function
      if node.id == 'int':
        return sympy.Function("int_")
      else:
        self.contain_unknown_func = True
        return sympy.Function(node.id)
    else:
      if node.attr in FUNCTION_MAPPING:
        return FUNCTION_MAPPING[node.attr]
      if node.attr == 'int':
        return sympy.Function("int_")
      else:
        names = self._get_attr_value(node, [])
        self.contain_unknown_func = True
        return sympy.Function('.'.join(names))

  def render_Call(self, node):
    if len(node.keywords):
      raise ValueError("Keyword arguments not supported.")
    elif getattr(node, 'starargs', None) is not None:
      raise ValueError("Variable number of arguments not supported")
    elif getattr(node, 'kwargs', None) is not None:
      raise ValueError("Keyword arguments not supported")
    elif len(node.args) == 0:
      return self.render_func(node.func)(sympy.Symbol('_placeholder_arg'))
    else:
      return self.render_func(node.func)(*(self.render_node(arg) for arg in node.args))

  def render_Compare(self, node):
    if len(node.comparators) > 1:
      raise SyntaxError("Can only handle single comparisons like a<b not a<b<c")
    op = node.ops[0]
    ops = self.expression_ops[op.__class__.__name__]
    left = self.render_node(node.left)
    right = self.render_node(node.comparators[0])
    return ops(left, right)

  def render_Name(self, node):
    if node.id in CONSTANT_MAPPING:
      return CONSTANT_MAPPING[node.id]
    else:
      return sympy.Symbol(node.id, real=True)

  def render_NameConstant(self, node):
    if node.value in [True, False]:
      return node.value
    else:
      return str(node.value)

  def render_Num(self, node):
    return sympy.Float(node.n)

  def render_BinOp(self, node):
    op_name = node.op.__class__.__name__
    # Sympy implements division and subtraction as multiplication/addition
    if op_name == 'Div':
      op = self.expression_ops['Mult']
      left = self.render_node(node.left)
      right = 1 / self.render_node(node.right)
      return op(left, right)
    elif op_name == 'FloorDiv':
      op = self.expression_ops['Mult']
      left = self.render_node(node.left)
      right = self.render_node(node.right)
      return sympy.floor(op(left, 1 / right))
    elif op_name == 'Sub':
      op = self.expression_ops['Add']
      left = self.render_node(node.left)
      right = -self.render_node(node.right)
      return op(left, right)
    else:
      op = self.expression_ops[op_name]
      left = self.render_node(node.left)
      right = self.render_node(node.right)
      return op(left, right)

  def render_BoolOp(self, node):
    op = self.expression_ops[node.op.__class__.__name__]
    return op(*(self.render_node(value) for value in node.values))

  def render_UnaryOp(self, node):
    op_name = node.op.__class__.__name__
    if op_name == 'UAdd':
      return self.render_node(node.operand)
    elif op_name == 'USub':
      return -self.render_node(node.operand)
    elif op_name == 'Not':
      return sympy.Not(self.render_node(node.operand))
    else:
      raise ValueError('Unknown unary operator: ' + op_name)


class Printer(StrPrinter):
  """
  Printer that overrides the printing of some basic sympy objects.

  e.g. print "a and b" instead of "And(a, b)".
  """

  def _print_And(self, expr):
    return ' and '.join(['(%s)' % self.doprint(arg) for arg in expr.args])

  def _print_Or(self, expr):
    return ' or '.join(['(%s)' % self.doprint(arg) for arg in expr.args])

  def _print_Not(self, expr):
    if len(expr.args) != 1:
      raise AssertionError('"Not" with %d arguments?' % len(expr.args))
    return f'not ({self.doprint(expr.args[0])})'

  def _print_Relational(self, expr):
    return f'{self.parenthesize(expr.lhs, precedence(expr))} ' \
           f'{self._relationals.get(expr.rel_op) or expr.rel_op} ' \
           f'{self.parenthesize(expr.rhs, precedence(expr))}'

  def _print_Function(self, expr):
    # Special workaround for the int function
    f_name = expr.func.__name__

    if f_name == 'Mod':
      return f'(({self.doprint(expr.args[0])})%({self.doprint(expr.args[1])}))'

    elif f_name in _func2expr:
      return f'{_func2expr[f_name]}({self.stringify(expr.args, ", ")})'

    else:
      return f"{f_name}({self.stringify(expr.args, ', ')})"


_PRINTER = Printer()


def str2sympy(str_expr):
  return Parser(str_expr)


def sympy2str(sympy_expr):
  # replace the standard functions by our names if necessary
  replacements = dict((f, sympy.Function(name)) for name, f in FUNCTION_MAPPING.items() if str(f) != name)

  # replace constants with our names as well
  replacements.update(dict((c, sympy.Symbol(name)) for name, c in CONSTANT_MAPPING.items() if str(c) != name))

  # Replace the placeholder argument by an empty symbol
  replacements[sympy.Symbol('_placeholder_arg')] = sympy.Symbol('')
  atoms = (sympy_expr.atoms() | {f.func for f in sympy_expr.atoms(sympy.Function)})
  for old, new in replacements.items():
    if old in atoms:
      sympy_expr = sympy_expr.subs(old, new)

  return _PRINTER.doprint(sympy_expr)


class Expression(object):
  def __init__(self, var, code):
    self.var_name = var
    self.code = code.strip()
    self.substituted_code = None

  @property
  def identifiers(self):
    return tools.get_identifiers(self.code)

  def __str__(self):
    return f'{self.var_name} = {self.code}'

  def __repr__(self):
    return self.__str__()

  def __eq__(self, other):
    if not isinstance(other, Expression):
      return NotImplemented
    if self.code != other.code:
      return False
    if self.var_name != other.var_name:
      return False
    return True

  def __ne__(self, other):
    return not self.__eq__(other)

  def get_code(self, subs=True):
    if subs:
      if self.substituted_code is None:
        return self.code
      else:
        return self.substituted_code
    else:
      return self.code


class SingleDiffEq(object):
  """Single Differential Equation.

  A differential equation is defined as the standard form:

  dx/dt = f(x) + g(x) dW

  Parameters
  ----------
  var_name : str
      The variable names.
  variables : list
      The code variables.
  expressions : list
      The code expressions for each line.
  derivative_expr : str
      The final derivative expression.
  scope : dict
      The code scope.
  """

  def __init__(self, var_name, variables, expressions, derivative_expr, scope, func_name):
    self.func_name = func_name
    # function scope
    self.func_scope = scope

    # differential variable name and time name
    self.var_name = var_name
    self.t_name = 't'

    # analyse function code
    self.expressions = [Expression(v, expr) for v, expr in zip(variables, expressions)]
    self.f_expr = Expression('_f_res_', derivative_expr)
    for k, num in Counter(variables).items():
      if num > 1:
        raise errors.AnalyzerError(
          f'Found "{k}" {num} times. Please assign each expression '
          f'in differential function with a unique name. ')

  def _substitute(self, final_exp, expressions, substitute_vars=None):
    """Substitute expressions to get the final single expression

    Parameters
    ----------
    final_exp : Expression
        The final expression.
    expressions : list, tuple
        The list/tuple of expressions.
    """
    if substitute_vars is None:
      return
    if final_exp is None:
      return
    assert substitute_vars == 'all' or \
           substitute_vars == self.var_name or \
           isinstance(substitute_vars, (tuple, list))

    # Goal: Substitute dependent variables into the expresion
    # Hint: This step doesn't require the left variables are unique
    dependencies = {}
    for expr in expressions:
      substitutions = {}
      for dep_var, dep_expr in dependencies.items():
        if dep_var in expr.identifiers:
          code = dep_expr.get_code(subs=True)
          substitutions[sympy.Symbol(dep_var, real=True)] = str2sympy(code).expr
      if len(substitutions):
        new_sympy_expr = str2sympy(expr.code).expr.xreplace(substitutions)
        new_str_expr = sympy2str(new_sympy_expr)
        expr.substituted_code = new_str_expr
        dependencies[expr.var_name] = expr
      else:
        if substitute_vars == 'all':
          dependencies[expr.var_name] = expr
        elif substitute_vars == self.var_name:
          if self.var_name in expr.identifiers:
            dependencies[expr.var_name] = expr
        else:
          ids = expr.identifiers
          for var in substitute_vars:
            if var in ids:
              dependencies[expr.var_name] = expr
              break

    # Goal: get the final differential equation
    # Hint: the step requires the expression variables must be unique
    substitutions = {}
    for dep_var, dep_expr in dependencies.items():
      code = dep_expr.get_code(subs=True)
      substitutions[sympy.Symbol(dep_var, real=True)] = str2sympy(code).expr
    if len(substitutions):
      new_sympy_expr = str2sympy(final_exp.code).expr.xreplace(substitutions)
      new_str_expr = sympy2str(new_sympy_expr)
      final_exp.substituted_code = new_str_expr

  def get_f_expressions(self, substitute_vars=None):
    if self.f_expr is None:
      return []
    self._substitute(self.f_expr, self.expressions, substitute_vars=substitute_vars)

    return_expressions = []
    # the derivative expression
    dif_eq_code = self.f_expr.get_code(subs=True)
    return_expressions.append(Expression(f'_df{self.var_name}_dt', dif_eq_code))
    # needed variables
    need_vars = tools.get_identifiers(dif_eq_code)
    # get the total return expressions
    for expr in self.expressions[::-1]:
      if expr.var_name in need_vars:
        if expr.substituted_code is None:
          code = expr.code
        else:
          code = expr.substituted_code
        return_expressions.append(Expression(expr.var_name, code))
        need_vars |= tools.get_identifiers(code)
    return return_expressions[::-1]

  def _replace_expressions(self, expressions, name, y_sub, t_sub=None):
    """Replace expressions of df part.

    Parameters
    ----------
    expressions : list, tuple
        The list/tuple of expressions.
    name : str
        The name of the new expression.
    y_sub : str
        The new name of the variable "y".
    t_sub : str, optional
        The new name of the variable "t".

    Returns
    -------
    list_of_expr : list
        A list of expressions.
    """
    return_expressions = []

    # replacements
    replacement = {self.var_name: y_sub}
    if t_sub is not None:
      replacement[self.t_name] = t_sub

    # replace variables in expressions
    for expr in expressions:
      replace = False
      identifiers = expr.identifiers
      for repl_var in replacement.keys():
        if repl_var in identifiers:
          replace = True
          break
      if replace:
        code = tools.word_replace(expr.code, replacement)
        new_expr = Expression(f"{expr.var_name}_{name}", code)
        return_expressions.append(new_expr)
        replacement[expr.var_name] = new_expr.var_name
    return return_expressions

  def replace_f_expressions(self, name, y_sub, t_sub=None):
    """Replace expressions of df part.

    Parameters
    ----------
    name : str
        The name of the new expression.
    y_sub : str
        The new name of the variable "y".
    t_sub : str, optional
        The new name of the variable "t".

    Returns
    -------
    list_of_expr : list
        A list of expressions.
    """
    return self._replace_expressions(self.get_f_expressions(),
                                     name=name,
                                     y_sub=y_sub,
                                     t_sub=t_sub)

  @property
  def expr_names(self):
    return [expr.var_name for expr in self.expressions]



# -*- coding: utf-8 -*-

from typing import Union, Sequence, Dict, Optional

import jax.numpy as jnp

import brainpy.math as bm
from brainpy.base.base import Base
from brainpy.base.collector import TensorCollector
from brainpy.errors import MathError
from brainpy.math.jaxarray import Variable
from .scheduler import make_schedule, Scheduler

__all__ = [
  'Optimizer',
  'SGD',
  'Momentum',
  'MomentumNesterov',
  'Adagrad',
  'Adadelta',
  'RMSProp',
  'Adam',
  'LARS',
]


class Optimizer(Base):
  """Base Optimizer Class.
  """

  def __init__(
      self,
      lr: Union[float, int, Scheduler],
      train_vars: Union[Sequence[Variable], Dict[str, Variable]] = None,
      name: str = None
  ):
    super(Optimizer, self).__init__(name=name)
    self.lr = make_schedule(lr)
    self.vars_to_train = TensorCollector()
    self.register_vars(train_vars)

  def register_vars(self, train_vars: Optional[Dict[str, Variable]] = None):
    raise NotImplementedError

  def check_grads(self, grads):
    if len(grads) != len(self.vars_to_train):
      raise MathError(
        f'The length of "grads" must be equal to "self.vars_to_train", '
        f'while we got {len(grads)} != {len(self.vars_to_train)}!')

  def __repr__(self):
    return f"{self.__class__.__name__}(lr={self.lr})"


class SGD(Optimizer):
  r"""Stochastic gradient descent optimizer.

  SGD performs a parameter update for training examples :math:`x` and label
  :math:`y`:

  .. math::

      \theta = \theta - \eta \cdot \nabla_\theta J(\theta; x; y)

  """

  def __init__(self, lr, train_vars=None, name=None):
    super(SGD, self).__init__(lr=lr, train_vars=train_vars, name=name)

  def register_vars(self, train_vars: Optional[Dict[str, Variable]] = None):
    train_vars = dict() if train_vars is None else train_vars
    if not isinstance(train_vars, dict):
      raise MathError('"train_vars" must be a dict of Variable.')
    self.vars_to_train.update(train_vars)

  def update(self, grads: dict):
    self.check_grads(grads)
    for key, p in self.vars_to_train.items():
      p.value -= self.lr() * grads[key]
    self.lr.update()


class Momentum(Optimizer):
  r"""Momentum optimizer.

  Momentum [1]_ is a method that helps accelerate SGD in the relevant direction
  and dampens oscillations. It does this by adding a fraction :math:`\gamma`
  of the update vector of the past time step to the current update vector:

  .. math::

    \begin{align}
    \begin{split}
    v_t &= \gamma v_{t-1} + \eta \nabla_\theta J( \theta) \\
    \theta &= \theta - v_t
    \end{split}
    \end{align}


  References
  ----------

  .. [1] Qian, N. (1999). On the momentum term in gradient descent learning
         algorithms. Neural Networks : The Official Journal of the International
         Neural Network Society, 12(1), 145–151. http://doi.org/10.1016/S0893-6080(98)00116-6

  """

  def __init__(self, lr, train_vars=None, momentum=0.9, name=None):
    super(Momentum, self).__init__(lr=lr, train_vars=train_vars, name=name)

    self.momentum = momentum

  def register_vars(self, train_vars: Optional[Dict[str, Variable]] = None):
    train_vars = dict() if train_vars is None else train_vars
    if not isinstance(train_vars, dict):
      raise MathError('"train_vars" must be a dict of Variable.')
    self.vars_to_train.update(train_vars)
    vs = dict((key + '_v', Variable(bm.zeros_like(x)))
              for key, x in train_vars.items())
    self.register_implicit_vars(vs)

  def update(self, grads: dict):
    self.check_grads(grads)
    lr = self.lr()
    for key, p in self.vars_to_train.items():
      g = grads[key]
      v = self.implicit_vars[key + '_v']
      v.value = self.momentum * v.value - lr * g
      p.value += v.value
    self.lr.update()

  def __repr__(self):
    return f"{self.__class__.__name__}(lr={self.lr}, momentum={self.momentum})"


class MomentumNesterov(Optimizer):
  r"""Nesterov accelerated gradient optimizer [2]_.

  .. math::

      \begin{align}
      \begin{split}
      v_t &= \gamma v_{t-1} + \eta \nabla_\theta J( \theta - \gamma v_{t-1} ) \\
      \theta &= \theta - v_t
      \end{split}
      \end{align}


  References
  ----------
  .. [2] Nesterov, Y. (1983). A method for unconstrained convex minimization problem with the rate of convergence o(1/k2). Doklady ANSSSR (translated as Soviet.Math.Docl.), vol. 269, pp. 543– 547.

  """

  def __init__(self, lr, train_vars=None, momentum=0.9, name=None):
    super(MomentumNesterov, self).__init__(lr=lr, train_vars=train_vars, name=name)

    self.momentum = momentum

  def register_vars(self, train_vars: Optional[Dict[str, Variable]] = None):
    train_vars = dict() if train_vars is None else train_vars
    if not isinstance(train_vars, dict):
      raise MathError('"train_vars" must be a dict of Variable.')
    self.vars_to_train.update(train_vars)
    vs = dict((key + '_v', Variable(bm.zeros_like(x)))
              for key, x in train_vars.items())
    self.register_implicit_vars(vs)

  def update(self, grads: dict):
    self.check_grads(grads)
    lr = self.lr()
    for key, p in self.vars_to_train.items():
      g = grads[key]
      v = self.implicit_vars[key + '_v']
      v.value = self.momentum * v.value - lr * g
      p.value += (self.momentum * v.value - lr * g)
    self.lr.update()

  def __repr__(self):
    return f"{self.__class__.__name__}(lr={self.lr}, momentum={self.momentum})"


class Adagrad(Optimizer):
  r"""Optimizer that implements the Adagrad algorithm.

  Adagrad [3]_ is an optimizer with parameter-specific learning rates, which are
  adapted relative to how frequently a parameter gets updated during training.
  The more updates a parameter receives, the smaller the updates.

  .. math::

      \theta_{t+1} = \theta_{t} - \dfrac{\eta}{\sqrt{G_{t} + \epsilon}} \odot g_{t}

  where :math:`G(t)` contains the sum of the squares of the past gradients

  One of Adagrad's main benefits is that it eliminates the need to manually tune
  the learning rate. Most implementations use a default value of 0.01 and leave it at that.
  Adagrad's main weakness is its accumulation of the squared gradients in the denominator:
  Since every added term is positive, the accumulated sum keeps growing during training.
  This in turn causes the learning rate to shrink and eventually become infinitesimally
  small, at which point the algorithm is no longer able to acquire additional knowledge.

  References
  ----------
  .. [3] Duchi, J., Hazan, E., & Singer, Y. (2011). Adaptive Subgradient Methods for Online Learning and Stochastic Optimization. Journal of Machine Learning Research, 12, 2121–2159. Retrieved from http://jmlr.org/papers/v12/duchi11a.html

  """

  def __init__(self, lr, train_vars=None, epsilon=1e-6, name=None):
    super(Adagrad, self).__init__(lr=lr, train_vars=train_vars, name=name)
    self.epsilon = epsilon

  def register_vars(self, train_vars: Optional[Dict[str, Variable]] = None):
    train_vars = dict() if train_vars is None else train_vars
    if not isinstance(train_vars, dict):
      raise MathError('"train_vars" must be a dict of Variable.')
    self.vars_to_train.update(train_vars)
    caches = dict((key + '_cache', Variable(bm.zeros_like(x)))
                  for key, x in train_vars.items())
    self.register_implicit_vars(caches)

  def update(self, grads: dict):
    self.check_grads(grads)
    lr = self.lr()
    for key, p in self.vars_to_train.items():
      g = grads[key]
      c = self.implicit_vars[key + '_cache']
      c.value += g ** 2
      p.value -= lr * g / bm.sqrt(c + self.epsilon)
    self.lr.update()

  def __repr__(self):
    return f"{self.__class__.__name__}(lr={self.lr}, epsilon={self.epsilon})"


class Adadelta(Optimizer):
  r"""Optimizer that implements the Adadelta algorithm.

  Adadelta [4]_ optimization is a stochastic gradient descent method that is based
  on adaptive learning rate per dimension to address two drawbacks:

  - The continual decay of learning rates throughout training.
  - The need for a manually selected global learning rate.

  Adadelta is a more robust extension of Adagrad that adapts learning rates based on
  a moving window of gradient updates, instead of accumulating all past gradients.
  This way, Adadelta continues learning even when many updates have been done. Compared
  to Adagrad, in the original version of Adadelta you don't have to set an initial
  learning rate.

  .. math::

    \boldsymbol{s}_t \leftarrow \rho \boldsymbol{s}_{t-1} + (1 - \rho) \boldsymbol{g}_t \odot \boldsymbol{g}_t, \\
    \boldsymbol{g}_t' \leftarrow \sqrt{\frac{\Delta\boldsymbol{x}_{t-1} + \epsilon}{\boldsymbol{s}_t + \epsilon}}   \odot \boldsymbol{g}_t, \\
    \boldsymbol{x}_t \leftarrow \boldsymbol{x}_{t-1} - \boldsymbol{g}'_t, \\
    \Delta\boldsymbol{x}_t \leftarrow \rho \Delta\boldsymbol{x}_{t-1} + (1 - \rho) \boldsymbol{g}'_t \odot \boldsymbol{g}'_t.

  :math:`\rho` should be between 0 and 1. A value of rho close to 1 will decay the
  moving average slowly and a value close to 0 will decay the moving average fast.

  :math:`\rho` = 0.95 and :math:`\epsilon`=1e-6 are suggested in the paper and reported
  to work for multiple datasets (MNIST, speech).

  In the paper, no learning rate is considered (so learning_rate=1.0). Probably best to
  keep it at this value. epsilon is important for the very first update (so the
  numerator does not become 0).

  References
  ----------
  .. [4] Zeiler, M. D. (2012). ADADELTA: An Adaptive Learning Rate Method. Retrieved from http://arxiv.org/abs/1212.5701

  """

  def __init__(self, train_vars=None, lr=0.01, epsilon=1e-6, rho=0.95, name=None):
    super(Adadelta, self).__init__(lr=lr, train_vars=train_vars, name=name)

    self.epsilon = epsilon
    self.rho = rho

  def register_vars(self, train_vars: Optional[Dict[str, Variable]] = None):
    train_vars = dict() if train_vars is None else train_vars
    if not isinstance(train_vars, dict):
      raise MathError('"train_vars" must be a dict of Variable.')
    self.vars_to_train.update(train_vars)
    caches = dict((key + '_cache', Variable(bm.zeros_like(x)))
                  for key, x in train_vars.items())
    deltas = dict((key + '_delta', Variable(bm.zeros_like(x)))
                  for key, x in train_vars.items())
    self.register_implicit_vars(caches)
    self.register_implicit_vars(deltas)

  def update(self, grads: dict):
    self.check_grads(grads)
    for key, p in self.vars_to_train.items():
      g = grads[key]
      c = self.implicit_vars[key + '_cache']
      d = self.implicit_vars[key + '_delta']
      c.value = self.rho * c.value + (1 - self.rho) * g ** 2
      update = g * jnp.sqrt(d.value + self.epsilon) / jnp.sqrt(c + self.epsilon)
      p.value -= update
      d.value = self.rho * d.value + (1 - self.rho) * update ** 2

  def __repr__(self):
    return (f"{self.__class__.__name__}(lr={self.lr}, "
            f"epsilon={self.epsilon}, rho={self.rho})")


class RMSProp(Optimizer):
  r"""Optimizer that implements the RMSprop algorithm.

  RMSprop [5]_ and Adadelta have both been developed independently around the same time
  stemming from the need to resolve Adagrad's radically diminishing learning rates.

  The gist of RMSprop is to:

  - Maintain a moving (discounted) average of the square of gradients
  - Divide the gradient by the root of this average

  .. math::

    \begin{split}c_t &= \rho c_{t-1} + (1-\rho)*g^2\\
    p_t &= \frac{\eta}{\sqrt{c_t + \epsilon}} * g \end{split}

  The centered version additionally maintains a moving average of the gradients,
  and uses that average to estimate the variance.

  References
  ----------
  .. [5] Tieleman, T. and Hinton, G. (2012):
         Neural Networks for Machine Learning, Lecture 6.5 - rmsprop.
         Coursera. http://www.youtube.com/watch?v=O3sxAc4hxZU (formula @5:20)
  """

  def __init__(self, lr, train_vars=None, epsilon=1e-6, rho=0.9, name=None):
    super(RMSProp, self).__init__(lr=lr, train_vars=train_vars, name=name)

    self.epsilon = epsilon
    self.rho = rho

  def register_vars(self, train_vars: Optional[Dict[str, Variable]] = None):
    train_vars = dict() if train_vars is None else train_vars
    if not isinstance(train_vars, dict):
      raise MathError('"train_vars" must be a dict of Variable.')
    self.vars_to_train.update(train_vars)
    caches = dict((key + '_cache', Variable(bm.zeros_like(x)))
                  for key, x in train_vars.items())
    self.register_implicit_vars(caches)

  def update(self, grads: dict):
    self.check_grads(grads)
    lr = self.lr()
    for k, p in self.vars_to_train.items():
      g = grads[k]
      c = self.implicit_vars[k + '_cache']
      c.value = self.rho * c.value + (1 - self.rho) * g ** 2
      p.value -= (lr * g / jnp.sqrt(c.value + self.epsilon))
    self.lr.update()

  def __repr__(self):
    return (f"{self.__class__.__name__}(lr={self.lr}, "
            f"epsilon={self.epsilon}, rho={self.rho})")


class Adam(Optimizer):
  """Optimizer that implements the Adam algorithm.

  Adam [6]_ - a stochastic gradient descent method (SGD) that computes
  individual adaptive learning rates for different parameters from estimates of
  first- and second-order moments of the gradients.

  Parameters
  ----------
  beta1: optional, float
    A positive scalar value for beta_1, the exponential decay rate
    for the first moment estimates (default 0.9).
  beta2: optional, float
    A positive scalar value for beta_2, the exponential decay rate
    for the second moment estimates (default 0.999).
  eps: optional, float
    A positive scalar value for epsilon, a small constant for
    numerical stability (default 1e-8).
  name : optional, str
    The optimizer name.

  References
  ----------
  .. [6] Kingma, D. P., & Ba, J. (2014). Adam: A method for stochastic optimization. arXiv preprint arXiv:1412.6980.
  """

  def __init__(self, lr, train_vars=None, beta1=0.9, beta2=0.999, eps=1e-8, name=None):
    super(Adam, self).__init__(lr=lr, train_vars=train_vars, name=name)

    self.beta1 = beta1
    self.beta2 = beta2
    self.eps = eps

  def __repr__(self):
    return (f"{self.__class__.__name__}(lr={self.lr}, "
            f"beta1={self.beta1}, beta2={self.beta2}, eps={self.eps})")

  def register_vars(self, train_vars: Optional[Dict[str, Variable]] = None):
    train_vars = dict() if train_vars is None else train_vars
    if not isinstance(train_vars, dict):
      raise MathError('"train_vars" must be a dict of Variable.')
    self.vars_to_train.update(train_vars)
    ms = dict((k + '_m', Variable(bm.zeros_like(x)))
              for k, x in train_vars.items())
    self.register_implicit_vars(ms)
    vs = dict((k + '_v', Variable(bm.zeros_like(x)))
              for k, x in train_vars.items())
    self.register_implicit_vars(vs)

  def update(self, grads: dict):
    self.check_grads(grads)
    lr = self.lr()
    lr /= (1 - self.beta1 ** (self.lr.step[0] + 1))
    lr *= jnp.sqrt(1 - self.beta2 ** (self.lr.step[0] + 1))
    for key, p in self.vars_to_train.items():
      m = self.implicit_vars[key + '_m']
      v = self.implicit_vars[key + '_v']
      g = grads[key]
      # First  moment estimate.
      m.value = self.beta1 * m.value + (1 - self.beta1) * g
      # Second moment estimate.
      v.value = self.beta2 * v.value + (1 - self.beta2) * g ** 2
      # Bias correction.
      p.value -= lr * m.value / (jnp.sqrt(v.value) + self.eps)
    self.lr.update()


class LARS(Optimizer):
  """Layer-wise adaptive rate scaling (LARS) optimizer.

  Parameters
  ----------
  momentum: float
    coefficient used for the moving average of the gradient.
  weight_decay: float
    weight decay coefficient.
  tc: float
    trust coefficient eta ( < 1) for trust ratio computation.
  eps: float
    epsilon used for trust ratio computation.
  """

  def __init__(self,
               lr: Union[float, int, Scheduler],
               train_vars: Dict[str, Variable] = None,
               momentum: float = 0.9,
               weight_decay: float = 1e-4,
               tc: float = 1e-3,
               eps: float = 1e-5,
               name: str = None):
    super(LARS, self).__init__(lr=lr, train_vars=train_vars, name=name)

    self.momentum = momentum
    self.weight_decay = weight_decay
    self.tc = tc
    self.eps = eps

  def __repr__(self):
    return (f"{self.__class__.__name__}(lr={self.lr}, "
            f"momentum={self.momentum}, weight_decay={self.weight_decay}, "
            f"tc={self.tc}, eps={self.eps})")

  def register_vars(self, train_vars: Optional[Dict[str, Variable]] = None):
    train_vars = dict() if train_vars is None else train_vars
    if not isinstance(train_vars, dict):
      raise MathError('"train_vars" must be a dict of Variable.')
    self.vars_to_train.update(train_vars)
    ms = dict((k + '_m', Variable(bm.zeros_like(x)))
              for k, x in train_vars.items())
    self.register_implicit_vars(ms)

  def update(self, grads: dict):
    self.check_grads(grads)
    lr = self.lr()
    for k, p in self.vars_to_train.items():
      g = grads[k]
      m = self.implicit_vars[k + '_m']
      p_norm = jnp.linalg.norm(bm.as_device_array(p))
      g_norm = jnp.linalg.norm(bm.as_device_array(g))
      trust_ratio = self.tc * p_norm / (g_norm + self.weight_decay * p_norm + self.eps)
      local_lr = lr * jnp.maximum(jnp.logical_or(p_norm == 0, g_norm == 0), trust_ratio)
      m.value = self.momentum * m.value + local_lr * (g + self.weight_decay * p.value)
      p.value -= m.value
    self.lr.update()



# -*- coding: utf-8 -*-

"""
This module provides numerical solvers for various differential equations,
including:

- ordinary differential equations (ODEs)
- stochastic differential equations (SDEs)
- delay differential equations (DDEs)
- fractional differential equations (FDEs)

Details please see the following.
"""

# basic tools
from .base import *
from .constants import *
from .joint_eq import *
from .runner import *

# ODE tools
from . import ode
from .ode.base import ODEIntegrator
from .ode.generic import (odeint,
                          get_default_odeint,
                          set_default_odeint,
                          register_ode_integrator)

# SDE tools
from . import sde
from .sde.base import SDEIntegrator
from .sde.generic import (sdeint,
                          get_default_sdeint,
                          set_default_sdeint,
                          register_sde_integrator)

# DDE tools
from . import dde
from .dde.base import DDEIntegrator
from .dde.generic import (ddeint,
                          get_default_ddeint,
                          set_default_ddeint,
                          register_dde_integrator)

# FDE tools
from . import fde
from .fde.base import FDEIntegrator
from .fde.generic import (fdeint,
                          get_default_fdeint,
                          set_default_fdeint,
                          register_fde_integrator)


# PDE tools
from . import pde



# -*- coding: utf-8 -*-

from typing import Union, Callable, Dict

import jax.numpy as jnp

import brainpy.math as bm
from brainpy.errors import UnsupportedError
from brainpy.integrators.base import Integrator
from brainpy.integrators.utils import get_args

__all__ = [
  'FDEIntegrator'
]


class FDEIntegrator(Integrator):
  """Numerical integrator for fractional differential equations (FEDs).

  Parameters
  ----------
  f : callable
    The derivative function.
  alpha: int, float, jnp.ndarray, bm.ndarray, sequence
    The fractional-order of the derivative function.
  dt: float, int
    The numerical precision.
  name: str
    The integrator name.
  """

  """The fraction order for each variable."""
  alpha: jnp.ndarray

  """The numerical integration precision."""
  dt: Union[float, int]

  """The fraction derivative function."""
  f: Callable

  def __init__(
      self,
      f: Callable,
      alpha,
      dt: float = None,
      name: str = None,
      state_delays: Dict[str, bm.AbstractDelay] = None,
  ):
    dt = bm.get_dt() if dt is None else dt
    parses = get_args(f)
    variables = parses[0]  # variable names, (before 't')
    parameters = parses[1]  # parameter names, (after 't')
    arguments = parses[2]  # function arguments

    # super initialization
    super(FDEIntegrator, self).__init__(name=name,
                                        variables=variables,
                                        parameters=parameters,
                                        arguments=arguments,
                                        dt=dt,
                                        state_delays=state_delays)

    # derivative function
    self.f = f

    # fractional-order
    if isinstance(alpha, (int, float)):
      alpha = jnp.ones(len(self.variables)) * alpha
    elif isinstance(alpha, (jnp.ndarray, bm.ndarray)):
      alpha = bm.as_device_array(alpha)
    elif isinstance(alpha, (list, tuple)):
      for a in alpha:
        assert isinstance(a, (float, int)), (f'Must be a tuple/list of int/float, '
                                             f'but we got {type(a)}: {a}')
      alpha = jnp.asarray(alpha)
    else:
      raise UnsupportedError(f'Do not support {type(alpha)}, please '
                             f'set fractional-order as number/tuple/list/tensor.')
    if len(alpha) != len(self.variables):
      raise ValueError(f'There are {len(self.variables)} variables, '
                       f'while we only got {len(alpha)} fractional-order '
                       f'settings: {alpha}')
    self.alpha = alpha

